# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

from .. import model

from ..model import expressions

from . import parser


def parse_type(source: str) -> model.Type:
    return parser.parse_type(parser.TokenStream(source))


def parse_expression(source: str) -> model.Expression:
    return parser.parse_expression(parser.TokenStream(source))


def parse(source: str, *, ctx: t.Optional[model.Context] = None) -> model.Context:
    return parser.parse_moml(parser.TokenStream(source), ctx=ctx)


def expr(source: str, **macros: expressions.ValueOrExpression) -> model.Expression:
    """Parses an expression in MOML syntax with macros."""
    return parser.parse_expression(
        parser.TokenStream(source.strip()),
        environment=parser.Environment(
            {name: expressions.ensure_expr(value) for name, value in macros.items()}
        ),
    )


def prop(source: str, **macros: expressions.ValueOrExpression) -> model.Expression:
    """Parses a property in MOML syntax with macros."""
    return parser.parse_property(
        parser.TokenStream(source.strip()),
        environment=parser.Environment(
            {name: expressions.ensure_expr(value) for name, value in macros.items()}
        ),
    )



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import enum
import inspect
import warnings

from .clstools import get_subclasses


class NoImplementationError(Exception):
    pass


class NoImplementationWarning(UserWarning):
    pass


def check_singledispatch(
    function: t.Callable[..., t.Any],
    superclass: type,
    *,
    error: bool = False,
    recursive: bool = True,
    ignore: t.AbstractSet[type] = frozenset(),
) -> None:
    if not hasattr(function, "registry") or not hasattr(function.registry, "keys"):
        raise ValueError(f"{function} is not a singledispatch function")
    subclasses = set(get_subclasses(superclass, recursive=recursive))
    subclasses -= ignore
    for cls in function.registry.keys():
        if not issubclass(cls, superclass):
            continue
        subclasses.discard(cls)
        if recursive:
            subclasses -= get_subclasses(cls, recursive=True)
    for subclass in subclasses:
        if inspect.isabstract(subclass):
            continue
        msg = (
            f"implementation of {function.__name__} for subclass {subclass} is missing"
        )
        if error:
            raise NoImplementationError(msg)
        else:
            warnings.warn(NoImplementationWarning(msg), stacklevel=2)


class NoEntryError(Exception):
    pass


class NoEntryWarning(UserWarning):
    pass


EnumType = t.TypeVar("EnumType", bound=enum.Enum)


def check_enum_map(
    enum_typ: t.Type[EnumType],
    mapping: t.Mapping[EnumType, t.Any],
    *,
    error: bool = False,
) -> None:
    for entry in enum_typ:
        if entry in mapping:
            continue
        msg = f"entry for {entry} is missing"
        if error:
            raise NoEntryError(msg)
        else:
            warnings.warn(NoEntryWarning(msg))



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

from . import checks

from .cache import lru_cache
from .clstools import get_subclasses


__all__ = ["checks", "lru_cache", "get_subclasses"]



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import dataclasses
import fractions

from .. import model
from ..model import actions, expressions, types, properties, operators

from . import lexer


_IGNORE = {lexer.TokenType.WHITESPACE, lexer.TokenType.COMMENT}


class MomlSyntaxError(ValueError):
    pass


Expected = t.Union[
    lexer.TokenType, str, t.AbstractSet[lexer.TokenType], t.AbstractSet[str]
]


_PRIMITIVE_TYPES = {
    "bool": types.BOOL,
    "int": types.INT,
    "real": types.REAL,
    "clock": types.CLOCK,
    "continuous": types.CONTINUOUS,
}


class TokenStream:
    code: str
    iterator: t.Iterator[lexer.Token]

    _current_token: t.Optional[lexer.Token]
    _next_token: t.Optional[lexer.Token]

    def __init__(self, code: str) -> None:
        self.code = code
        self.iterator = lexer.lex(code)
        self._current_token = self._forward()
        self._next_token = self._forward()

    def _forward(self) -> t.Optional[lexer.Token]:
        try:
            token = next(self.iterator)
            while token.token_type in _IGNORE:
                token = next(self.iterator)
            return token
        except StopIteration:
            return None

    def consume(self) -> lexer.Token:
        if self._current_token is None:
            raise ValueError("no token to consume")
        token = self._current_token
        self._current_token = self._next_token
        self._next_token = self._forward()
        return token

    def expect(self, expected: Expected, *, consume: bool = True) -> lexer.Token:
        token = self.accept(expected, consume=consume)
        if token is None:
            raise self.make_error(f"expected {expected}")
        return token

    def accept(
        self, expected: Expected, *, consume: bool = True
    ) -> t.Optional[lexer.Token]:
        current_token = self._current_token
        if current_token is None:
            return None
        if isinstance(expected, lexer.TokenType):
            if current_token.token_type is not expected:
                return None
        elif isinstance(expected, str):
            if current_token.text != expected:
                return None
        elif (
            current_token.token_type not in expected
            and current_token.text not in expected
        ):
            return None
        if consume:
            return self.consume()
        else:
            return current_token

    def check(self, expected: Expected) -> bool:
        return bool(self.accept(expected, consume=False))

    @property
    def finished_parsing(self) -> bool:
        return self._current_token is None

    @property
    def current_token(self) -> lexer.Token:
        if self._current_token is None:
            raise self.make_error("expected token but found EOF")
        return self._current_token

    def make_error(self, message: str) -> MomlSyntaxError:
        start_row = max(0, self.current_token.start_row - 3)
        end_row = self.current_token.start_row + 1
        rows = "\n".join(self.code.split("\n")[start_row:end_row])
        return MomlSyntaxError(
            f"{message} at {self.current_token.start_row}:{self.current_token.start_column} "
            f"(token: `{self.current_token.text}`)\n\n"
            f"{rows}\n"
            f"{' ' * self.current_token.start_column}^"
        )


def parse_primitive_type(stream: TokenStream) -> types.Type:
    return _PRIMITIVE_TYPES[stream.expect(_PRIMITIVE_TYPES.keys()).text]


def parse_type(stream: TokenStream) -> types.Type:
    typ = parse_primitive_type(stream)
    while stream.accept("["):
        if stream.accept("]"):
            typ = types.ArrayType(typ)
        else:
            if not isinstance(typ, types.NumericType):
                raise stream.make_error(
                    f"{typ} is not numeric; only numeric types can be bounded"
                )
            lower_bound = parse_expression(stream)
            stream.expect(",")
            upper_bound = parse_expression(stream)
            stream.expect("]")
            typ = typ.bound(lower_bound, upper_bound)
    return typ


_PRECEDENCE = {
    lexer.TokenType.POWER: 41,
    lexer.TokenType.STAR: 31,
    lexer.TokenType.SLASH: 31,
    lexer.TokenType.SLASH_SLASH: 31,
    lexer.TokenType.PERCENTAGE: 31,
    lexer.TokenType.PLUS: 30,
    lexer.TokenType.MINUS: 30,
    lexer.TokenType.COMP_EQ: 20,
    lexer.TokenType.COMP_NEQ: 20,
    lexer.TokenType.COMP_LT: 20,
    lexer.TokenType.COMP_LE: 20,
    lexer.TokenType.COMP_GE: 20,
    lexer.TokenType.COMP_GT: 20,
    lexer.TokenType.LOGIC_AND: 12,
    lexer.TokenType.LOGIC_OR: 11,
    lexer.TokenType.LOGIC_XOR: 11,
    lexer.TokenType.LOGIC_IMPLIES: 10,
    lexer.TokenType.LOGIC_EQUIV: 10,
    lexer.TokenType.QUESTIONMARK: 5,
}

_RIGHT_ASSOCIATIVE = {lexer.TokenType.LOGIC_IMPLIES}

_BINARY_CONSTRUCTORS: t.Mapping[lexer.TokenType, expressions.BinaryConstructor] = {
    lexer.TokenType.POWER: expressions.power,
    lexer.TokenType.STAR: expressions.mul,
    lexer.TokenType.SLASH: expressions.real_div,
    lexer.TokenType.SLASH_SLASH: expressions.floor_div,
    lexer.TokenType.PERCENTAGE: expressions.mod,
    lexer.TokenType.PLUS: expressions.add,
    lexer.TokenType.MINUS: expressions.sub,
    lexer.TokenType.COMP_EQ: expressions.equals,
    lexer.TokenType.COMP_NEQ: expressions.not_equals,
    lexer.TokenType.COMP_LT: expressions.less,
    lexer.TokenType.COMP_LE: expressions.less_or_equal,
    lexer.TokenType.COMP_GE: expressions.greater_or_equal,
    lexer.TokenType.COMP_GT: expressions.greater,
    lexer.TokenType.LOGIC_AND: expressions.logic_and,
    lexer.TokenType.LOGIC_OR: expressions.logic_or,
    lexer.TokenType.LOGIC_XOR: expressions.logic_xor,
    lexer.TokenType.LOGIC_IMPLIES: expressions.logic_implies,
    lexer.TokenType.LOGIC_EQUIV: expressions.logic_equiv,
}


_BuiltinExpressionConstructor = t.Callable[[t.List[model.Expression]], model.Expression]
_BuiltinPropertyConstructor = t.Callable[[t.List[model.Expression]], model.Expression]


@d.dataclass
class Environment:
    macro_expressions: t.Dict[str, model.Expression] = d.field(default_factory=dict)


def _construct_floor(arguments: t.List[model.Expression]) -> model.Expression:
    if len(arguments) != 1:
        raise Exception(
            f"floor takes exactly 1 argument but {len(arguments)} are given"
        )
    return expressions.floor(arguments[0])


def _construct_ceil(arguments: t.List[model.Expression]) -> model.Expression:
    if len(arguments) != 1:
        raise Exception(f"ceil takes exactly 1 argument but {len(arguments)} are given")
    return expressions.ceil(arguments[0])


def _construct_abs(arguments: t.List[model.Expression]) -> model.Expression:
    if len(arguments) != 1:
        raise Exception(f"abs takes exactly 1 argument but {len(arguments)} are given")
    return expressions.absolute(arguments[0])


def _construct_sgn(arguments: t.List[model.Expression]) -> model.Expression:
    if len(arguments) != 1:
        raise Exception(f"sgn takes exactly 1 argument but {len(arguments)} are given")
    return expressions.sgn(arguments[0])


def _construct_trc(arguments: t.List[model.Expression]) -> model.Expression:
    if len(arguments) != 1:
        raise Exception(f"sgn takes exactly 1 argument but {len(arguments)} are given")
    return expressions.trunc(arguments[0])


def _construct_min(arguments: t.List[model.Expression]) -> model.Expression:
    if len(arguments) != 2:
        raise Exception(f"min takes exactly 2 argument but {len(arguments)} are given")
    return expressions.minimum(arguments[0], arguments[1])


def _construct_max(arguments: t.List[model.Expression]) -> model.Expression:
    if len(arguments) != 2:
        raise Exception(f"max takes exactly 2 argument but {len(arguments)} are given")
    return expressions.maximum(arguments[0], arguments[1])


def _construct_probability_min(arguments: t.List[model.Expression]) -> model.Expression:
    if len(arguments) != 1:
        raise Exception(f"Pmin takes exactly 1 argument but {len(arguments)} are given")
    return properties.min_prob(arguments[0])


def _construct_probability_max(arguments: t.List[model.Expression]) -> model.Expression:
    if len(arguments) != 1:
        raise Exception(f"Pmax takes exactly 1 argument but {len(arguments)} are given")
    return properties.max_prob(arguments[0])


def _construct_finally(
    arguments: t.List[model.Expression],
) -> model.Expression:
    if len(arguments) != 1:
        raise Exception(f"F takes exactly 1 argument but {len(arguments)} are given")
    return properties.eventually(arguments[0])


def _construct_globally(
    arguments: t.List[model.Expression],
) -> model.Expression:
    if len(arguments) != 1:
        raise Exception(f"G takes exactly 1 argument but {len(arguments)} are given")
    return properties.globally(arguments[0])


_BUILTIN_FUNCTIONS: t.Dict[str, _BuiltinExpressionConstructor] = {
    "floor": _construct_floor,
    "ceil": _construct_ceil,
    "abs": _construct_abs,
    "sgn": _construct_sgn,
    "trc": _construct_trc,
    "min": _construct_min,
    "max": _construct_max,
}


def _update_builtin_functions() -> None:
    def _make_constructor(
        function: operators.TrigonometricFunction,
    ) -> _BuiltinExpressionConstructor:
        def construct(arguments: t.List[model.Expression]) -> model.Expression:
            if len(arguments) != 1:
                symbol = function.symbol
                raise Exception(
                    f"{symbol} takes exactly 1 argument but {len(arguments)} are given"
                )
            return expressions.Trigonometric(function, arguments[0])

        return construct

    for trigonometric_function in operators.TrigonometricFunction:
        construct = _make_constructor(trigonometric_function)
        _BUILTIN_FUNCTIONS[trigonometric_function.symbol] = construct


_update_builtin_functions()


_PROPERTY_FUNCTIONS: t.Mapping[str, _BuiltinPropertyConstructor] = {
    "Pmin": _construct_probability_min,
    "Pmax": _construct_probability_max,
    "F": _construct_finally,
    "G": _construct_globally,
}

_AGGREGATION_FUNCTIONS: t.Mapping[str, operators.AggregationFunction] = {
    "exists": operators.AggregationFunction.EXISTS,
    "forall": operators.AggregationFunction.FORALL,
    **{operator.symbol: operator for operator in operators.AggregationFunction},
}


def _parse_property_function(
    constructor: _BuiltinPropertyConstructor,
    stream: TokenStream,
    environment: Environment,
) -> model.Expression:
    arguments: t.List[model.Expression] = []
    if not stream.accept(lexer.TokenType.RIGHT_PAR):
        arguments.append(parse_property(stream, environment=environment))
        while not stream.accept(lexer.TokenType.RIGHT_PAR):
            stream.expect(",")
            arguments.append(parse_property(stream, environment=environment))
    return constructor(arguments)


def _parse_builtin_function(
    constructor: _BuiltinExpressionConstructor,
    stream: TokenStream,
    environment: Environment,
) -> model.Expression:
    arguments: t.List[model.Expression] = []
    if not stream.accept(lexer.TokenType.RIGHT_PAR):
        arguments.append(parse_expression(stream, environment=environment))
        while not stream.accept(lexer.TokenType.RIGHT_PAR):
            stream.expect(",")
            arguments.append(parse_expression(stream, environment=environment))
    return constructor(arguments)


def _parse_primary(stream: TokenStream, environment: Environment) -> model.Expression:
    if stream.accept(lexer.TokenType.INTEGER, consume=False):
        return expressions.ensure_expr(int(stream.consume().text))
    elif stream.accept(lexer.TokenType.REAL, consume=False):
        return expressions.ensure_expr(fractions.Fraction(stream.consume().text))
    elif stream.accept("true"):
        return expressions.TRUE
    elif stream.accept("false"):
        return expressions.FALSE
    elif stream.accept(lexer.TokenType.MACRO, consume=False):
        name = stream.consume().match["macro_name"]
        try:
            return environment.macro_expressions[name]
        except KeyError:
            raise stream.make_error(f"no macro with name {name} found")
        if stream.accept("("):
            raise stream.make_error(
                "functional syntax-aware macros are not supported yet"
            )
    elif stream.accept(lexer.TokenType.IDENTIFIER, consume=False):
        identifier = stream.consume().text
        if stream.accept(lexer.TokenType.LEFT_PAR):
            if identifier == "sample":
                arguments: t.List[model.Expression] = []
                distribution = model.DistributionType.by_name(
                    stream.expect(lexer.TokenType.STRING).match["string"]
                )
                while not stream.accept(lexer.TokenType.RIGHT_PAR):
                    stream.expect(",")
                    arguments.append(parse_expression(stream, environment=environment))
                return expressions.Sample(distribution, tuple(arguments))
            elif identifier in _PROPERTY_FUNCTIONS:
                return _parse_property_function(
                    _PROPERTY_FUNCTIONS[identifier], stream, environment
                )
            elif identifier in _BUILTIN_FUNCTIONS:
                return _parse_builtin_function(
                    _BUILTIN_FUNCTIONS[identifier], stream, environment
                )
            raise stream.make_error(f"unknown function {identifier}")
        elif stream.accept(lexer.TokenType.FILTER_LEFT):
            function = _AGGREGATION_FUNCTIONS[identifier]
            values = parse_property(stream, environment=environment)
            stream.expect(lexer.TokenType.PIPE)
            if stream.accept("initial"):
                states = properties.INITIAL_STATES
            elif stream.accept("deadlock"):
                states = properties.DEADLOCK_STATES
            else:
                stream.expect("timelock")
                states = properties.TIMELOCK_STATES
            return properties.aggregate(function, values, states)
        else:
            return expressions.name(identifier)
    elif stream.accept(lexer.TokenType.LEFT_PAR):
        expr = parse_expression(stream, environment=environment)
        stream.expect(lexer.TokenType.RIGHT_PAR)
        return expr
    else:
        raise stream.make_error("expected primary expression or property")


def _parse_unary(stream: TokenStream, environment: Environment) -> model.Expression:
    if stream.accept(lexer.TokenType.LOGIC_NOT):
        operand = _parse_unary(stream, environment=environment)
        # FIXME: proper error reporting
        assert isinstance(operand, model.Expression)
        return expressions.logic_not(operand)
    elif stream.accept(lexer.TokenType.MINUS):
        operand = _parse_unary(stream, environment=environment)
        # FIXME: proper error reporting
        assert isinstance(operand, model.Expression)
        return expressions.sub(expressions.ensure_expr(0), operand)
    return _parse_primary(stream, environment=environment)


DEFAULT_ENVIRONMENT = Environment()


def parse_property(
    stream: TokenStream,
    *,
    min_precedence: int = 0,
    environment: Environment = DEFAULT_ENVIRONMENT,
) -> model.Expression:
    left = _parse_unary(stream, environment=environment)
    while _PRECEDENCE.get(stream.current_token.token_type, -1) >= min_precedence:
        operator = stream.consume()
        precedence = _PRECEDENCE[operator.token_type]
        if operator.token_type is lexer.TokenType.QUESTIONMARK:
            consequence = parse_expression(stream, environment=environment)
            stream.expect(":")
            alternative = parse_expression(
                stream, min_precedence=precedence, environment=environment
            )
            # FIXME: proper error reporting
            assert isinstance(left, model.Expression)
            left = expressions.ite(left, consequence, alternative)
        else:
            if operator.token_type not in _RIGHT_ASSOCIATIVE:
                precedence += 1
            right = parse_expression(
                stream, min_precedence=precedence, environment=environment
            )
            # FIXME: proper error reporting
            assert isinstance(left, model.Expression)
            left = _BINARY_CONSTRUCTORS[operator.token_type](left, right)
    return left


def parse_expression(
    stream: TokenStream,
    *,
    min_precedence: int = 0,
    environment: Environment = DEFAULT_ENVIRONMENT,
) -> model.Expression:
    prop = parse_property(
        stream, min_precedence=min_precedence, environment=environment
    )
    # FIXME: proper error reporting
    assert isinstance(prop, model.Expression)
    return prop


@dataclasses.dataclass(frozen=True)
class _DeclarationInfo:
    name: str
    typ: types.Type
    value: t.Optional[model.Expression] = None


def _parse_identifier_declaration(stream: TokenStream) -> _DeclarationInfo:
    name = stream.expect(lexer.TokenType.IDENTIFIER).text
    stream.expect(":")
    typ = parse_type(stream)
    if stream.accept(":="):
        return _DeclarationInfo(name, typ, parse_expression(stream))
    else:
        return _DeclarationInfo(name, typ)


def _parse_constant_declaration(stream: TokenStream) -> model.ConstantDeclaration:
    stream.expect("constant")
    info = _parse_identifier_declaration(stream)
    comment = stream.accept(lexer.TokenType.STRING)
    if comment is not None:
        return model.ConstantDeclaration(
            info.name, info.typ, value=info.value, comment=comment.match["string"]
        )
    return model.ConstantDeclaration(info.name, info.typ, value=info.value)


def _parse_variable_declaration(stream: TokenStream) -> model.VariableDeclaration:
    is_transient = bool(stream.accept("transient"))
    stream.expect("variable")
    info = _parse_identifier_declaration(stream)
    stream.accept(lexer.TokenType.STRING)  # TODO: comment
    return model.VariableDeclaration(
        info.name,
        info.typ,
        is_transient=is_transient,
        initial_value=info.value,
    )


def _parse_assignment(stream: TokenStream) -> model.Assignment:
    stream.expect("assign")
    name = stream.expect(lexer.TokenType.IDENTIFIER).text
    if stream.check(lexer.TokenType.INTEGER):
        index = int(stream.consume().text)
        if index < 0:
            stream.make_error("unexpected token")
    else:
        index = 0
    stream.expect(":=")
    value = parse_expression(stream)
    return model.Assignment(expressions.Name(name), value=value, index=index)


def _parse_location(stream: TokenStream, automaton: model.Automaton) -> model.Location:
    is_initial = bool(stream.accept("initial"))
    stream.expect("location")
    name = stream.expect(lexer.TokenType.IDENTIFIER).text
    progress_invariant: t.Optional[model.Expression] = None
    transient_values: t.Set[model.Assignment] = set()
    if stream.accept(":"):
        stream.expect(lexer.TokenType.INDENT)
        while not stream.accept(lexer.TokenType.DEDENT):
            if stream.accept("invariant"):
                if progress_invariant is not None:
                    raise stream.make_error("duplicate definition of invariant")
                progress_invariant = parse_expression(stream)
            elif stream.check("assign"):
                transient_values.add(_parse_assignment(stream))
            else:
                raise stream.make_error("expected location body")
    return automaton.create_location(
        name,
        progress_invariant=progress_invariant,
        transient_values=transient_values,
        initial=is_initial,
    )


def parse_automaton(stream: TokenStream, ctx: model.Context) -> model.Automaton:
    stream.expect("automaton")
    name = stream.expect(lexer.TokenType.IDENTIFIER).text
    automaton = ctx.create_automaton(name=name)
    if stream.accept("("):
        while not stream.accept(")"):
            parameter_declaration = _parse_identifier_declaration(stream)
            automaton.declare_parameter(
                parameter_declaration.name,
                parameter_declaration.typ,
                default_value=parameter_declaration.value,
            )
            if not stream.accept(","):
                stream.expect(")")
                break
    stream.expect(":")
    stream.expect(lexer.TokenType.INDENT)
    location_map: t.Dict[str, model.Location] = {}
    while not stream.accept(lexer.TokenType.DEDENT):
        if stream.check({"transient", "variable"}):
            declaration = _parse_variable_declaration(stream)
            automaton.scope.add_declaration(declaration)
        elif stream.check({"initial", "location"}):
            location = _parse_location(stream, automaton)
            assert location.name not in location_map and location.name is not None
            location_map[location.name] = location
        elif stream.accept("edge"):
            stream.expect("from")
            location_name = stream.expect(lexer.TokenType.IDENTIFIER).text
            stream.expect(":")
            stream.expect(lexer.TokenType.INDENT)
            action_pattern: t.Optional[model.ActionPattern] = None
            guard: t.Optional[model.Expression] = None
            rate: t.Optional[model.Expression] = None
            destinations: t.Set[model.Destination] = set()
            while not stream.accept(lexer.TokenType.DEDENT):
                if stream.accept("action"):
                    assert action_pattern is None
                    action_pattern = _parse_action_pattern(stream, ctx)
                elif stream.accept("guard"):
                    assert guard is None
                    guard = parse_expression(stream)
                elif stream.accept("rate"):
                    assert rate is None
                    rate = parse_expression(stream)
                elif stream.accept("to"):
                    target_name = stream.expect(lexer.TokenType.IDENTIFIER).text
                    probability: t.Optional[model.Expression] = None
                    assignments: t.Set[model.Assignment] = set()
                    if stream.accept(":"):
                        stream.expect(lexer.TokenType.INDENT)
                        while not stream.accept(lexer.TokenType.DEDENT):
                            if stream.accept("probability"):
                                assert probability is None
                                probability = parse_expression(stream)
                            elif stream.check("assign"):
                                assignments.add(_parse_assignment(stream))
                            else:
                                raise stream.make_error("unexpected token")
                    destinations.add(
                        model.Destination(
                            location=location_map[target_name],
                            probability=probability,
                            assignments=tuple(assignments),
                        )
                    )
                else:
                    raise stream.make_error("unexpected token")
            assert destinations, "missign destinations"
            automaton.create_edge(
                location_map[location_name],
                tuple(destinations),
                action_pattern=action_pattern,
                guard=guard,
                rate=rate,
            )
        else:
            raise stream.make_error("unexpected token")
    return automaton


def _parse_action_pattern(
    stream: TokenStream, ctx: model.Context
) -> t.Optional[model.ActionPattern]:
    if stream.accept({"-", "τ"}):
        return None
    action_type = ctx.get_action_type_by_name(
        stream.expect(lexer.TokenType.IDENTIFIER).text
    )
    arguments: t.List[actions.ActionArgument] = []
    if stream.accept("(") and not stream.accept(")"):
        while True:
            if stream.accept("<!"):
                arguments.append(actions.WriteArgument(parse_expression(stream)))
            elif stream.accept("?>"):
                arguments.append(
                    actions.ReadArgument(stream.expect(lexer.TokenType.IDENTIFIER).text)
                )
            else:
                arguments.append(
                    actions.GuardArgument(
                        stream.expect(lexer.TokenType.IDENTIFIER).text
                    )
                )
            if stream.accept(")"):
                break
            stream.expect(",")
    return model.ActionPattern(action_type, arguments=tuple(arguments))


def _parse_network(stream: TokenStream, ctx: model.Context) -> None:
    stream.expect("network")
    identifier = stream.accept(lexer.TokenType.IDENTIFIER)
    name: t.Optional[str] = None
    if identifier:
        name = identifier.text
    network = ctx.create_network(name=name)
    stream.expect(":")
    stream.expect(lexer.TokenType.INDENT)
    instance_map: t.Dict[str, model.Instance] = {}
    while not stream.accept(lexer.TokenType.DEDENT):
        if stream.accept("restrict"):
            stream.expect("initial")
            network.initial_restriction = parse_expression(stream)
        elif stream.accept("instance"):
            name = stream.expect(lexer.TokenType.IDENTIFIER).text
            assert name not in instance_map
            automaton_name = stream.expect(lexer.TokenType.IDENTIFIER).text
            parameters: t.List[expressions.Expression] = []
            if stream.accept("("):
                while not stream.accept(")"):
                    parameters.append(parse_expression(stream))
                    if not stream.accept(","):
                        stream.expect(")")
                        break
            input_enable: t.Set[actions.ActionType] = set()
            if stream.accept(":"):
                stream.expect(lexer.TokenType.INDENT)
                if stream.accept("input"):
                    stream.expect("enable")
                    input_enable.add(
                        ctx.get_action_type_by_name(
                            stream.expect(lexer.TokenType.IDENTIFIER).text
                        )
                    )
                    while stream.accept(","):
                        input_enable.add(
                            ctx.get_action_type_by_name(
                                stream.expect(lexer.TokenType.IDENTIFIER).text
                            )
                        )
                stream.expect(lexer.TokenType.DEDENT)
            instance_map[name] = ctx.get_automaton_by_name(
                automaton_name
            ).create_instance(arguments=parameters, input_enable=input_enable)
        elif stream.accept("composition"):
            instances: t.List[model.Instance] = []
            instances.append(
                instance_map[stream.expect(lexer.TokenType.IDENTIFIER).text]
            )
            while stream.accept("|"):
                instances.append(
                    instance_map[stream.expect(lexer.TokenType.IDENTIFIER).text]
                )
            if stream.accept(":"):
                stream.expect(lexer.TokenType.INDENT)
                while not stream.accept(lexer.TokenType.DEDENT):
                    stream.expect("synchronize")
                    parenthesis = stream.accept("(")
                    patterns: t.List[t.Optional[model.ActionPattern]] = []
                    patterns.append(_parse_action_pattern(stream, ctx))
                    while stream.accept("|"):
                        patterns.append(_parse_action_pattern(stream, ctx))
                    assert len(patterns) == len(instances)
                    stream.expect(lexer.TokenType.ARROW)
                    result_pattern = _parse_action_pattern(stream, ctx)
                    if parenthesis:
                        stream.expect(")")
                    vector: t.Dict[model.Instance, model.ActionPattern] = {}
                    for index, action_pattern in enumerate(patterns):
                        if action_pattern is not None:
                            vector[instances[index]] = action_pattern
                    network.create_link(vector, result=result_pattern)
        else:
            raise stream.make_error("expected network element")


def _parse_metadata(stream: TokenStream) -> t.Mapping[str, str]:
    stream.expect("metadata")
    stream.expect(":")
    stream.expect(lexer.TokenType.INDENT)
    fields: t.Dict[str, str] = {}
    while not stream.accept(lexer.TokenType.DEDENT):
        field_name = stream.expect(lexer.TokenType.STRING).match["string"]
        stream.expect(":")
        field_value = stream.expect(lexer.TokenType.STRING).match["string"]
        fields[field_name] = field_value
    return fields


def _parse_action_declaration(stream: TokenStream) -> model.ActionType:
    stream.expect("action")
    name = stream.expect(lexer.TokenType.IDENTIFIER).text
    comment: t.Optional[str] = None
    parameters: t.List[model.ActionParameter] = []
    if stream.accept("(") and not stream.accept(")"):
        while True:
            typ = parse_type(stream)
            parameter_comment: t.Optional[str] = None
            if stream.check(lexer.TokenType.STRING):
                parameter_comment = stream.consume().match["string"]
            parameters.append(model.ActionParameter(typ, comment=parameter_comment))
            if stream.accept(")"):
                break
            stream.expect(",")
    if stream.check(lexer.TokenType.STRING):
        comment = stream.consume().match["string"]
    return model.ActionType(name, tuple(parameters), comment=comment)


def _parse_moml(stream: TokenStream, ctx: model.Context) -> model.Context:
    automaton_map: t.Dict[str, model.Automaton] = {}
    while True:
        if stream.check({"transient", "variable"}):
            ctx.global_scope.add_declaration(_parse_variable_declaration(stream))
        elif stream.check("constant"):
            ctx.global_scope.add_declaration(_parse_constant_declaration(stream))
        elif stream.check("automaton"):
            automaton = parse_automaton(stream, ctx)
            assert automaton.name not in automaton_map and automaton.name is not None
            automaton_map[automaton.name] = automaton
        elif stream.check("action"):
            ctx._add_action_type(_parse_action_declaration(stream))
        elif stream.check("network"):
            _parse_network(stream, ctx)
        elif stream.check(lexer.TokenType.END_OF_FILE):
            break
        elif stream.check("metadata"):
            ctx.update_metadata(_parse_metadata(stream))
        else:
            raise stream.make_error("unexpected token")
    return ctx


def parse_moml(
    stream: TokenStream, *, ctx: t.Optional[model.Context] = None
) -> model.Context:
    if stream.accept("model_type"):
        model_type = model.ModelType[stream.expect(lexer.TokenType.IDENTIFIER).text]
    return _parse_moml(stream, ctx or model.Context(model_type))



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

from importlib import metadata


version = metadata.version(__package__)



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020-2021, Saarland University
# Copyright (C) 2020-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

from . import time, zones

from .explore import Action, Destination, Transition, State, Explorer, Parameters
from .time import TimeType, DiscreteTime
from .values import Value


__all__ = [
    "Action",
    "Destination",
    "Transition",
    "State",
    "Explorer",
    "Parameters",
    "TimeType",
    "DiscreteTime",
    "Value",
    "time",
    "zones",
]



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

from . import analysis, engine, jani, model, moml

from .metadata import version


__version__ = version


__all__ = ["analysis", "engine", "jani", "model", "moml"]



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020-2021, Saarland University
# Copyright (C) 2020-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>
# Copyright (C) 2020-2021, Michaela Klauck <klauck@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import enum

from . import errors, expressions, operators, types

if t.TYPE_CHECKING:
    from . import context


@d.dataclass(frozen=True)
class Aggregate(expressions.Expression):
    """
    Applies an aggregation function over a set of states.

    Attributes
    ----------
    function:
        The aggregation function to apply.
    values:
        The values to aggregate over.
    predcate:
        The predicate used to identify the states to aggregate over.
    """

    function: operators.AggregationFunction
    values: expressions.Expression
    predicate: expressions.Expression

    def infer_type(self, scope: context.Scope) -> types.Type:
        predicate_type = self.predicate.infer_type(scope)
        if not predicate_type == types.BOOL:
            raise errors.InvalidTypeError(
                f"expected types.BOOL but got {predicate_type}"
            )
        values_type = self.values.infer_type(scope)
        if values_type not in self.function.allowed_values_type:
            raise errors.InvalidTypeError(
                f"invalid type {values_type} of values in filter function"
            )
        return self.function.infer_result_type(values_type)

    @property
    def children(self) -> t.Sequence[expressions.Expression]:
        return (self.predicate,)


class StatePredicate(enum.Enum):
    """
    An enum of state predicates to be used with :class:`Aggregate`.
    """

    INITIAL = "initial"
    """ The state is an initial state. """

    DEADLOCK = "deadlock"
    """ The state is a deadlock state. """

    TIMELOCK = "timelock"
    """ The state is a timelock state. """


@d.dataclass(frozen=True)
class StateSelector(expressions.Expression):
    """
    State selector expression using :class:`StatePredicate`.

    Attributes
    ----------
    predicate:
        A :class:`StatePredicate`.
    """

    predicate: StatePredicate

    def infer_type(self, scope: context.Scope) -> types.Type:
        return types.BOOL

    @property
    def children(self) -> t.Sequence[expressions.Expression]:
        return ()


INITIAL_STATES = StateSelector(StatePredicate.INITIAL)
DEADLOCK_STATES = StateSelector(StatePredicate.DEADLOCK)
TIMELOCK_STATES = StateSelector(StatePredicate.TIMELOCK)


@d.dataclass(frozen=True)
class Probability(expressions.Expression):
    """
    Probability property.

    Attributes
    ----------
    operator:
        *Min* or *max* probability (:class:`~momba.model.operators.MinMax`).
    formula:
        Boolean expression to compute the probability for.
    """

    operator: operators.MinMax
    formula: expressions.Expression

    def infer_type(self, scope: context.Scope) -> types.Type:
        formula_type = self.formula.infer_type(scope)
        if not formula_type == types.BOOL:
            raise errors.InvalidTypeError(f"expected types.BOOL but got {formula_type}")
        return types.REAL

    @property
    def children(self) -> t.Sequence[expressions.Expression]:
        return (self.formula,)


@d.dataclass(frozen=True)
class PathQuantifier(expressions.Expression):
    """
    A temporal path quantifier property.

    Attributes
    ----------
    quantifier:
        The quantifier (:class:`~momba.model.operators.Quantifier`).
    formula:
        The inner formula.
    """

    quantifier: operators.Quantifier
    formula: expressions.Expression

    def infer_type(self, scope: context.Scope) -> types.Type:
        formula_type = self.formula.infer_type(scope)
        if not formula_type == types.BOOL:
            raise errors.InvalidTypeError(f"expected types.BOOL but got {formula_type}")
        return types.BOOL

    @property
    def children(self) -> t.Sequence[expressions.Expression]:
        return (self.formula,)


class AccumulationInstant(enum.Enum):
    """
    En enumeration of reward accumulation instants.
    """

    STEPS = "steps"
    """ Accumulate at each step. """

    TIME = "time"
    """ Accumulate with the progression of time. """

    EXIT = "exit"
    """ Accumulate after exiting a state. """


@d.dataclass(frozen=True)
class ExpectedReward(expressions.Expression):
    """
    Expected reward property.

    Attributes
    ----------
    operator:
        *Min* or *max* probability (:class:`~momba.model.operators.MinMax`).
    reward:
        Expression to compute the reward.
    accumulate:
        A set of accumulation instants.
    reachability:
    step_instant:
    time_instant:
    reward_instants:
    """

    operator: operators.MinMax
    reward: expressions.Expression
    accumulate: t.Optional[t.FrozenSet[AccumulationInstant]] = None
    reachability: t.Optional[expressions.Expression] = None
    step_instant: t.Optional[expressions.Expression] = None
    time_instant: t.Optional[expressions.Expression] = None
    reward_instants: t.Optional[t.Sequence[RewardInstant]] = None

    def infer_type(self, scope: context.Scope) -> types.Type:
        # TODO: check the types of the provided arguments
        return types.REAL

    @property
    def children(self) -> t.Sequence[expressions.Expression]:
        children: t.List[expressions.Expression] = []
        if self.reachability is not None:
            children.append(self.reachability)
        if self.step_instant is not None:
            children.append(self.step_instant)
        if self.time_instant is not None:
            children.append(self.time_instant)
        if self.reward_instants is not None:
            for reward_instant in self.reward_instants:
                children.extend(reward_instant.children)
        return children


@d.dataclass(frozen=True)
class RewardInstant:
    """
    A reward instant.

    Attributes
    ----------
    expression:
    accumulate:
    instant:
    """

    expression: expressions.Expression
    accumulate: t.FrozenSet[AccumulationInstant]
    instant: expressions.Expression

    @property
    def children(self) -> t.Sequence[expressions.Expression]:
        return (self.expression, self.instant)


@d.dataclass(frozen=True)
class SteadyState(expressions.Expression):
    """
    A *steady-state* property.

    Attributes
    ----------
    operator:
    formula:
    accumulate:
    """

    operator: operators.MinMax
    formula: expressions.Expression
    accumulate: t.Optional[t.FrozenSet[AccumulationInstant]] = None

    def infer_type(self, scope: context.Scope) -> types.Type:
        # TODO: check the types of the provided arguments
        return types.REAL

    @property
    def children(self) -> t.Sequence[expressions.Expression]:
        return (self.formula,)


@d.dataclass(frozen=True)
class BinaryPathFormula(expressions.Expression):
    """
    A temporal binary path formula.

    Attributes
    ----------
    operator:
    left:
    right:
    step_bounds:
    time_bounds:
    reward_bounds:
    """

    operator: operators.BinaryPathOperator
    left: expressions.Expression
    right: expressions.Expression
    step_bounds: t.Optional[Interval] = None
    time_bounds: t.Optional[Interval] = None
    reward_bounds: t.Optional[t.Sequence[RewardBound]] = None

    def infer_type(self, scope: context.Scope) -> types.Type:
        left_type = self.left.infer_type(scope)
        if left_type != types.BOOL:
            raise errors.InvalidTypeError(f"expected types.BOOL but got {left_type}")
        right_type = self.left.infer_type(scope)
        if right_type != types.BOOL:
            raise errors.InvalidTypeError(f"expected types.BOOL but got {right_type}")
        # TODO: check the types of the other arguments
        return types.BOOL

    @property
    def children(self) -> t.Sequence[expressions.Expression]:
        children: t.List[expressions.Expression] = [self.left, self.right]
        if self.step_bounds is not None:
            children.extend(self.step_bounds.expressions)
        if self.time_bounds is not None:
            children.extend(self.time_bounds.expressions)
        if self.reward_bounds is not None:
            for reward_bound in self.reward_bounds:
                children.extend(reward_bound.expressions)
        return children


@d.dataclass(frozen=True)
class UnaryPathFormula(expressions.Expression):
    """
    A temporal unary path formula.

    Attributes
    ----------
    operator:
    formula:
    step_bounds:
    time_bounds:
    reward_bounds:
    """

    operator: operators.UnaryPathOperator
    formula: expressions.Expression
    step_bounds: t.Optional[Interval] = None
    time_bounds: t.Optional[Interval] = None
    reward_bounds: t.Optional[t.Sequence[RewardBound]] = None

    def infer_type(self, scope: context.Scope) -> types.Type:
        formula_type = self.formula.infer_type(scope)
        if formula_type != types.BOOL:
            raise errors.InvalidTypeError(f"expected types.BOOL but got {formula_type}")
        # TODO: check the types of the other arguments
        return types.BOOL

    @property
    def children(self) -> t.Sequence[expressions.Expression]:
        children: t.List[expressions.Expression] = [self.formula]
        if self.step_bounds is not None:
            children.extend(self.step_bounds.expressions)
        if self.time_bounds is not None:
            children.extend(self.time_bounds.expressions)
        if self.reward_bounds is not None:
            for reward_bound in self.reward_bounds:
                children.extend(reward_bound.expressions)
        return children


@d.dataclass(frozen=True)
class Interval:
    """
    An interval.

    Attributes
    ----------
    lower:
        The lower bound of the interval or :code:`None`.
    upper:
        The upper bound of the interval or :code:`None`.
    lower_exclusive:
        Whether the lower bound is exclusive.
    upper_exclusive:
        Whether the upper bound is exclusive.
    """

    lower: t.Optional[expressions.Expression] = None
    upper: t.Optional[expressions.Expression] = None
    lower_exclusive: t.Optional[expressions.Expression] = None
    upper_exclusive: t.Optional[expressions.Expression] = None

    @property
    def expressions(self) -> t.Sequence[expressions.Expression]:
        return [
            expr
            for expr in [
                self.lower,
                self.upper,
                self.lower_exclusive,
                self.upper_exclusive,
            ]
            if expr is not None
        ]


@d.dataclass(frozen=True)
class RewardBound:
    """
    A *reward bound*.

    Attributes
    ----------
    expression:
    accumulate:
    bounds:
    """

    expression: expressions.Expression
    accumulate: t.FrozenSet[AccumulationInstant]
    bounds: Interval

    @property
    def expressions(self) -> t.Sequence[expressions.Expression]:
        expressions = [self.expression]
        expressions.extend(self.bounds.expressions)
        return expressions


def aggregate(
    function: operators.AggregationFunction,
    values: expressions.Expression,
    states: expressions.Expression = INITIAL_STATES,
) -> expressions.Expression:
    """
    Creates an :class:`Aggregate` property.
    """
    return Aggregate(function, values, states)


def min_prob(formula: expressions.Expression) -> expressions.Expression:
    """
    Constructs a :math:`P_\\mathit{min}` property.
    """
    return Probability(operators.MinMax.MIN, formula)


def max_prob(formula: expressions.Expression) -> expressions.Expression:
    """
    Constructs a :math:`P_\\mathit{max}` property.
    """
    return Probability(operators.MinMax.MAX, formula)


def forall_paths(formula: expressions.Expression) -> expressions.Expression:
    """
    CTL :math:`\\forall` path operator.
    """
    return PathQuantifier(operators.Quantifier.FORALL, formula)


def exists_path(formula: expressions.Expression) -> expressions.Expression:
    """
    CTL :math:`\\exists` exists operator.
    """
    return PathQuantifier(operators.Quantifier.EXISTS, formula)


def min_expected_reward(
    reward: expressions.Expression,
    *,
    accumulate: t.Optional[t.AbstractSet[AccumulationInstant]] = None,
    reachability: t.Optional[expressions.Expression] = None,
    step_instant: t.Optional[expressions.Expression] = None,
    time_instant: t.Optional[expressions.Expression] = None,
    reward_instants: t.Optional[t.Sequence[RewardInstant]] = None,
) -> expressions.Expression:
    """
    Constructs a :math:`E_\\mathit{min}` property.
    """
    return ExpectedReward(
        operators.MinMax.MIN,
        reward,
        accumulate=None if accumulate is None else frozenset(accumulate),
        reachability=reachability,
        step_instant=step_instant,
        time_instant=time_instant,
        reward_instants=reward_instants,
    )


def max_expected_reward(
    reward: expressions.Expression,
    *,
    accumulate: t.Optional[t.AbstractSet[AccumulationInstant]] = None,
    reachability: t.Optional[expressions.Expression] = None,
    step_instant: t.Optional[expressions.Expression] = None,
    time_instant: t.Optional[expressions.Expression] = None,
    reward_instants: t.Optional[t.Sequence[RewardInstant]] = None,
) -> expressions.Expression:
    """
    Constructs a :math:`E_\\mathit{max}` property.
    """
    return ExpectedReward(
        operators.MinMax.MAX,
        reward,
        accumulate=None if accumulate is None else frozenset(accumulate),
        reachability=reachability,
        step_instant=step_instant,
        time_instant=time_instant,
        reward_instants=reward_instants,
    )


def min_steady_state(
    formula: expressions.Expression,
    *,
    accumulate: t.Optional[t.AbstractSet[AccumulationInstant]] = None,
) -> expressions.Expression:
    """
    Constructs a :math:`S_\\mathit{min}` property.
    """
    return SteadyState(
        operators.MinMax.MIN,
        formula,
        accumulate=None if accumulate is None else frozenset(accumulate),
    )


def max_steady_state(
    formula: expressions.Expression,
    *,
    accumulate: t.Optional[t.AbstractSet[AccumulationInstant]] = None,
) -> expressions.Expression:
    """
    Constructs a :math:`S_\\mathit{max}` property.
    """
    return SteadyState(
        operators.MinMax.MAX,
        formula,
        accumulate=None if accumulate is None else frozenset(accumulate),
    )


def until(
    left: expressions.Expression,
    right: expressions.Expression,
    *,
    step_bounds: t.Optional[Interval] = None,
    time_bounds: t.Optional[Interval] = None,
    reward_bounds: t.Optional[t.Sequence[RewardBound]] = None,
) -> expressions.Expression:
    """
    Constructs a temporal *until* property.
    """
    return BinaryPathFormula(
        operators.BinaryPathOperator.UNTIL,
        left,
        right,
        step_bounds=step_bounds,
        time_bounds=time_bounds,
        reward_bounds=reward_bounds,
    )


def weak_until(
    left: expressions.Expression,
    right: expressions.Expression,
    *,
    step_bounds: t.Optional[Interval] = None,
    time_bounds: t.Optional[Interval] = None,
    reward_bounds: t.Optional[t.Sequence[RewardBound]] = None,
) -> expressions.Expression:
    """
    Constructs a temporal *weak-until* property.
    """
    return BinaryPathFormula(
        operators.BinaryPathOperator.WEAK_UNTIL,
        left,
        right,
        step_bounds=step_bounds,
        time_bounds=time_bounds,
        reward_bounds=reward_bounds,
    )


def release(
    left: expressions.Expression,
    right: expressions.Expression,
    *,
    step_bounds: t.Optional[Interval] = None,
    time_bounds: t.Optional[Interval] = None,
    reward_bounds: t.Optional[t.Sequence[RewardBound]] = None,
) -> expressions.Expression:
    """
    Constructs a temporal *release* property.
    """
    return BinaryPathFormula(
        operators.BinaryPathOperator.RELEASE,
        left,
        right,
        step_bounds=step_bounds,
        time_bounds=time_bounds,
        reward_bounds=reward_bounds,
    )


def eventually(
    formula: expressions.Expression,
    *,
    step_bounds: t.Optional[Interval] = None,
    time_bounds: t.Optional[Interval] = None,
    reward_bounds: t.Optional[t.Sequence[RewardBound]] = None,
) -> expressions.Expression:
    """
    Constructs a temporal *evenutally* property.
    """
    return UnaryPathFormula(
        operators.UnaryPathOperator.EVENTUALLY,
        formula,
        step_bounds=step_bounds,
        time_bounds=time_bounds,
        reward_bounds=reward_bounds,
    )


def globally(
    formula: expressions.Expression,
    *,
    step_bounds: t.Optional[Interval] = None,
    time_bounds: t.Optional[Interval] = None,
    reward_bounds: t.Optional[t.Sequence[RewardBound]] = None,
) -> expressions.Expression:
    """
    Constructs a temporal *globally* property.
    """
    return UnaryPathFormula(
        operators.UnaryPathOperator.GLOBALLY,
        formula,
        step_bounds=step_bounds,
        time_bounds=time_bounds,
        reward_bounds=reward_bounds,
    )



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import abc

from . import errors

if t.TYPE_CHECKING:
    from . import context, expressions, types


@d.dataclass(frozen=True)
class ActionParameter:
    """
    Represents an action parameter.

    Attributes
    ----------
    typ:
        The type of the parameter.
    comment:
        An optional comment for the parameter.
    """

    typ: types.Type
    comment: t.Optional[str] = None


@d.dataclass(frozen=True)
class ActionType:
    """
    Represents an action type.

    Attributes
    ----------
    label:
        The label of the action type.
    parameters:
        The parameters of the action type.
    comment:
        An optional comment for the action type.
    """

    label: str
    parameters: t.Tuple[ActionParameter, ...] = ()
    comment: t.Optional[str] = None

    @property
    def has_parameters(self) -> bool:
        """
        Returns :obj:`True` if and only if the action type has parameters.
        """
        return bool(self.parameters)

    @property
    def arity(self) -> int:
        """
        The arity, i.e., number of parameters, of the action type.
        """
        return len(self.parameters)

    def create_pattern(self, *arguments: ActionArgument) -> ActionPattern:
        return ActionPattern(self, arguments=arguments)


class ActionArgument(abc.ABC):
    """
    Represents an argument for an action pattern.
    """


@d.dataclass(frozen=True)
class WriteArgument(ActionArgument):
    """
    Writes the value of the expression for the respective parameter.
    """

    # is evaluated in the automata's scope, not in the edge scope
    expression: expressions.Expression


@d.dataclass(frozen=True)
class ReadArgument(ActionArgument):
    """
    Reads a value from the respective parameter.
    """

    # gets declared in the edge scope
    identifier: str


@d.dataclass(frozen=True)
class GuardArgument(ActionArgument):
    """
    Represents an argument to be used in a guard.
    """

    identifier: str


@d.dataclass(frozen=True)
class ActionPattern:
    """
    Represents an action pattern.

    Attributes
    ----------
    action_type:
        The type of the action pattern.
    arguments:
        The arguments of the action pattern.
    """

    action_type: ActionType
    arguments: t.Tuple[ActionArgument, ...] = ()

    def __post_init__(self) -> None:
        if len(self.action_type.parameters) != len(self.arguments):
            raise errors.InvalidTypeError(
                f"action type {self.action_type} requires {self.action_type.arity} "
                f"arguments but {len(self.arguments)} were given"
            )

    def declare_in(self, scope: context.Scope) -> None:
        """
        Declares the identifiers of the pattern in the given scope.

        .. warning::
            Experimental feature used for value passing only.
        """
        for argument, parameter in zip(self.arguments, self.action_type.parameters):
            if isinstance(argument, GuardArgument):
                if scope.is_declared(argument.identifier):
                    assert scope.lookup(argument.identifier).typ == parameter.typ
                else:
                    scope.declare_variable(
                        argument.identifier, parameter.typ, is_transient=True
                    )
            elif isinstance(argument, ReadArgument):
                scope.declare_variable(
                    argument.identifier, parameter.typ, is_transient=True
                )



# -*- coding:utf-8 -*-
#
# Copyright (C) 2021, Saarland University
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import json

try:
    import torch  # type: ignore
except ImportError:
    raise ImportError(
        "Missing optional dependency `torch`.\n"
        "Using MombaGym's NN export function requires `torch`."
    )


class UnsupportedLayerError(Exception):
    pass


class _FakeTensor:
    parent: t.Optional[_FakeTensor]
    module: t.Optional[torch.nn.Module]

    def __init__(
        self,
        parent: t.Optional[_FakeTensor] = None,
        module: t.Optional[torch.nn.Module] = None,
    ):
        self.parent = parent
        self.module = module

    def __torch_function__(self, func, types, args=(), kwargs=None):  # type: ignore
        if func is torch.functional.F.linear:
            (_, weight) = args
            bias = kwargs.get("bias", None)
            out_features, in_features = weight.shape
            module = torch.nn.Linear(in_features, out_features, bias=bias is not None)
            module.weight = weight
            if bias is not None:
                module.bias = bias
            return _FakeTensor(self, module)
        elif func is torch.functional.F.relu:
            return _FakeTensor(self, torch.nn.ReLU())
        elif func is torch.functional.F.celu:
            return _FakeTensor(self, torch.nn.CELU(alpha=kwargs.get("alpha", 1.0)))
        else:
            raise UnsupportedLayerError(
                f"unsupported function {func!r} applied to layer"
            )


def _dump_layer(name: str, layer: torch.nn.Module) -> t.Any:
    result: t.Dict[str, t.Any] = {"name": name, "kind": layer.__class__.__name__}
    if isinstance(layer, torch.nn.Linear):
        result["inputSize"] = layer.in_features
        result["outputSize"] = layer.out_features
        result["hasBiases"] = getattr(layer, "bias", None) is not None
        result["weights"] = layer.weight.tolist()
        result["biases"] = layer.bias.tolist()
    elif isinstance(layer, torch.nn.ReLU):
        pass
    elif isinstance(layer, torch.nn.CELU):
        result["alpha"] = layer.alpha
    else:
        raise UnsupportedLayerError(f"layer of type {type(layer)} not supported")
    return result


def _dump_layers(net: torch.nn.Sequential) -> t.Any:
    return [_dump_layer(name, layer) for name, layer in net.named_children()]


def dump_nn(net: torch.nn.Module) -> str:
    if not isinstance(net, torch.nn.Sequential):
        result = net.forward(_FakeTensor())
        assert isinstance(
            result, _FakeTensor
        ), "the result of forwarding should be a `_FakeTensor`"
        layers = []
        while result.module is not None:
            layers.append(result.module)
            result = result.parent
        net = torch.nn.Sequential(*reversed(layers))
    return json.dumps({"layers": _dump_layers(net)}, indent=2)



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import abc

from . import errors, expressions

if t.TYPE_CHECKING:
    from . import context


class BaseTypeError(errors.TypeConstructionError):
    pass


class InvalidBoundError(errors.TypeConstructionError):
    pass


Bound = t.Optional[t.Union["expressions.ValueOrExpression", "ellipsis"]]  # noqa:


class Type(abc.ABC):
    """
    A *type*.
    """

    @property
    def is_numeric(self) -> bool:
        """
        Indicates whether the type is numeric or not.
        """
        return False

    @abc.abstractmethod
    def is_assignable_from(self, typ: Type) -> bool:
        """
        Checks whether values of the other type can be assigned to
        targets of this type.
        """
        raise NotImplementedError()

    def validate_in(self, scope: context.Scope) -> None:  # noqa: B027
        pass


class NumericType(Type, abc.ABC):
    @property
    def is_numeric(self) -> bool:
        return True

    def bound(self, lower: Bound, upper: Bound) -> BoundedType:
        """
        Bounds the numeric type with the given bounds.
        """
        return BoundedType(
            self, BoundedType.convert_bound(lower), BoundedType.convert_bound(upper)
        )


@d.dataclass(frozen=True)
class IntegerType(NumericType):
    """
    The *integer type*.

    .. hint::
        Instead of instantiating new objects of this class use :attr:`INT`.
    """

    def __str__(self) -> str:
        return "types.INT"

    def is_assignable_from(self, typ: Type) -> bool:
        if isinstance(typ, BoundedType):
            return typ.base == INT
        return typ == INT


@d.dataclass(frozen=True)
class RealType(NumericType):
    """
    The *real type*.

    .. hint::
        Instead of instantiating new objects of this class use :attr:`REAL`.
    """

    def __str__(self) -> str:
        return "types.REAL"

    def is_assignable_from(self, typ: Type) -> bool:
        return typ.is_numeric


@d.dataclass(frozen=True)
class BoolType(Type):
    """
    The *boolean type*.

    .. hint::
        Instead of instantiating new objects of this class use :attr:`BOOL`.
    """

    def __str__(self) -> str:
        return "types.BOOL"

    def is_assignable_from(self, typ: Type) -> bool:
        return typ == BOOL


@d.dataclass(frozen=True)
class ClockType(NumericType):
    """
    The *clock type*.

    .. hint::
        Instead of instantiating new objects of this class use :attr:`CLOCK`.
    """

    def __str__(self) -> str:
        return "types.CLOCK"

    def is_assignable_from(self, typ: Type) -> bool:
        if isinstance(typ, BoundedType):
            return typ.base == INT
        return typ == INT


@d.dataclass(frozen=True)
class ContinuousType(NumericType):
    """
    The *continuous type*.

    .. hint::
        Instead of instantiating new objects of this class use :attr:`CONTINUOUS`.
    """

    def __str__(self) -> str:
        return "types.CONTINUOUS"

    def is_assignable_from(self, typ: Type) -> bool:
        return typ.is_numeric


INT = IntegerType()
REAL = RealType()
BOOL = BoolType()
CLOCK = ClockType()
CONTINUOUS = ContinuousType()


@d.dataclass(frozen=True)
class BoundedType(NumericType):
    """
    A *bounded numeric type*.

    Attributes
    ----------
    base:
        The numeric base type.
    lower_bound:
        The optional lower bound for values of the type.
    upper_bound:
        The optional upper bound for values of the type.
    """

    base: NumericType

    lower_bound: t.Optional[expressions.Expression]
    upper_bound: t.Optional[expressions.Expression]

    def __str__(self) -> str:
        return f"{self.base}[{self.lower_bound}, {self.upper_bound}]"

    def __post_init__(self) -> None:
        if not isinstance(self.base, NumericType):
            raise BaseTypeError("base-type of bounded type must be numeric")
        if self.lower_bound is None and self.upper_bound is None:
            raise InvalidBoundError(
                "neither `lower_bound` nor `upper_bound` is present"
            )

    def validate_in(self, scope: context.Scope) -> None:
        if self.lower_bound is not None:
            # if not scope.is_constant(self.lower_bound):
            #     raise InvalidBoundError("`lower_bound` has to be a constant")
            lower_bound_type = scope.get_type(self.lower_bound)
            if not self.base.is_assignable_from(lower_bound_type):
                raise InvalidBoundError(
                    f"type {lower_bound_type} of `lower_bound` is not "
                    f"assignable to base-type {self.base}"
                )
        if self.upper_bound is not None:
            # if not scope.is_constant(self.upper_bound):
            #     raise InvalidBoundError("`upper_bound` has to be a constant")
            upper_bound_type = scope.get_type(self.upper_bound)
            if not self.base.is_assignable_from(upper_bound_type):
                raise InvalidBoundError(
                    f"type {upper_bound_type} of `upper_bound` is not "
                    f"assignable to base-type {self.base}"
                )

    @staticmethod
    def convert_bound(bound: Bound) -> t.Optional[expressions.Expression]:
        if bound is None or bound is Ellipsis:
            return None
        return expressions.ensure_expr(t.cast(expressions.ValueOrExpression, bound))

    def is_assignable_from(self, typ: Type) -> bool:
        return self.base.is_assignable_from(typ)


@d.dataclass(frozen=True)
class ArrayType(Type):
    """
    An *array type*.

    Attributes
    ----------
    element:
        The type of the elements of the array.
    """

    element: Type

    def __str__(self) -> str:
        return f"Array({self.element})"

    def is_assignable_from(self, typ: Type) -> bool:
        return isinstance(typ, ArrayType) and self.element.is_assignable_from(
            typ.element
        )


def array_of(element: Type) -> ArrayType:
    """
    Constructs an array type with the given element type.
    """
    return ArrayType(element)


@d.dataclass(frozen=True)
class SetType(Type):
    """
    An *set type*.

    Attributes
    ----------
    element:
        The type of the elements of the set.
    """

    element: Type

    def __str__(self) -> str:
        return f"Set({self.element})"

    def is_assignable_from(self, typ: Type) -> bool:
        return isinstance(typ, SetType) and self.element.is_assignable_from(typ.element)


def set_of(element: Type) -> SetType:
    """
    Constructs a set type with the given element type.
    """
    return SetType(element)


@d.dataclass(frozen=True)
class StateType(Type):
    def __str__(self) -> str:
        return "types.STATE"

    def is_assignable_from(self, typ: Type) -> bool:
        return typ == STATE


STATE = StateType()


class InferTypeUnary(t.Protocol):
    def __call__(self, operand: Type) -> Type:
        pass


class InferTypeBinary(t.Protocol):
    def __call__(self, left: Type, right: Type) -> Type:
        pass



# -*- coding:utf-8 -*-
#
# Copyright (C) 2021, Saarland University
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

from . import errors, expressions, types

if t.TYPE_CHECKING:
    from . import context


@d.dataclass(frozen=True)
class FunctionParameter:
    """
    A *function parameter*.

    Attributes
    ----------
    name:
        The name of the parameter.
    typ:
        The type of the parameter.
    """

    name: str
    typ: types.Type


@d.dataclass(frozen=True)
class FunctionDefinition:
    """
    A *function definition*.

    Attributes
    ----------
    name:
        The name of the defined function.
    parameters:
        The parameters of the function.
    returns:
        The return type of the function.
    body:
        The body of the function.
    """

    name: str
    parameters: t.Tuple[FunctionParameter, ...]
    returns: types.Type
    body: expressions.Expression


@d.dataclass(frozen=True)
class CallExpression(expressions.Expression):
    """
    A *call expression*.

    Attributes
    ----------
    function:
        The name of the function to be called.
    arguments:
        The arguments to be passed to the function.
    """

    function: str
    arguments: t.Tuple[expressions.Expression, ...]

    def infer_type(self, scope: context.Scope) -> types.Type:
        definition = scope.get_function(self.function)
        if len(self.arguments) != len(definition.parameters):
            raise errors.InvalidTypeError(
                f"function {self.function} expects {len(definition.parameters)}"
                f" arguments but {len(self.arguments)} are given"
            )
        for argument, parameter in zip(self.arguments, definition.parameters):
            argument_type = argument.infer_type(scope)
            if not parameter.typ.is_assignable_from(argument_type):
                raise errors.InvalidTypeError(
                    f"parameter {parameter.name} of function {self.function}"
                    f" has type {parameter.typ} but argument of type"
                    f" {argument_type} has been provided"
                )
        return definition.returns

    @property
    def children(self) -> t.Sequence[expressions.Expression]:
        return self.arguments



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

import dataclasses as d
import typing as t

import fractions
import json
import pathlib
import subprocess
import sys
import tempfile


from .. import model
from ..analysis import checkers
from ..jani import dump_model

from .errors import ToolError, ToolTimeoutError


Timeout = t.Optional[t.Union[float, int]]
Command = t.Sequence[t.Union[str, pathlib.Path]]


_MODEST_URL = "https://www.modestchecker.net/Downloads/"

_USER_HOME = pathlib.Path("~").expanduser()

_is_windows = sys.platform == "win32"
_is_macos = sys.platform == "darwin"


if _is_windows:
    _DATA_PATH = _USER_HOME / "Application Data" / "koehlma" / "Momba"
    _MODEST_PLATFORM = "win"
elif _is_macos:
    _DATA_PATH = _USER_HOME / "Library" / "Application Support" / "Momba"
    _MODEST_PLATFORM = "osx"
else:
    _DATA_PATH = _USER_HOME / ".local" / "share" / "momba"
    _MODEST_PLATFORM = "linux"


_MODEST_PATH = _DATA_PATH / "modest"

if _is_windows:
    _MODEST_EXE = _MODEST_PATH / "Modest" / "modest.exe"
else:
    _MODEST_EXE = _MODEST_PATH / "Modest" / "modest"


_FILE_REGEX = (
    r"""<a href="(?P<file>Modest-Toolset-(?P<version>[v.0-9\-a-z]+?)"""
    r"""-(?P<platform>[a-z]+)-x64\.zip)"""
)


def _install_locally() -> None:
    import os
    import re
    import stat
    import zipfile

    from urllib import request

    response = request.urlopen(_MODEST_URL)
    if response.status != 200:
        raise Exception(f"unable to download Modest (error {response.status})")
    page = response.read().decode("utf-8")
    for match in re.finditer(_FILE_REGEX, page):
        if match["platform"] != _MODEST_PLATFORM:
            continue
        response = request.urlopen(f"{_MODEST_URL}{match['file']}")
        if response.status != 200:
            raise Exception(f"unable to download Modest (error {response.status})")
        with tempfile.TemporaryDirectory(prefix="momba") as temp_directory:
            zip_path = pathlib.Path(temp_directory) / "modest.zip"
            zip_path.write_bytes(response.read())
            with zipfile.ZipFile(zip_path) as zip_file:
                zip_file.extractall(_MODEST_PATH)
        os.chmod(_MODEST_EXE, os.stat(_MODEST_EXE).st_mode | stat.S_IEXEC)
        break
    else:
        raise Exception(f"unable to download Modest for platform {_MODEST_PLATFORM}")


def _setup_locally() -> None:
    if not _MODEST_EXE.exists():
        _install_locally()


@d.dataclass(eq=False)
class Toolset:
    """Interface to the Modest Toolset."""

    executable: t.Union[str, pathlib.Path]
    """Path to the executable of the Modest Toolset."""

    environment: t.Optional[t.Mapping[str, str]] = None
    """Environment variables for the execution."""

    def check(
        self,
        arguments: t.Iterable[t.Union[str, int, float, pathlib.Path]],
        *,
        timeout: Timeout = None,
        capture_output: bool = True,
    ) -> t.Mapping[str, t.Any]:
        """Runs `modest check` with the provided arguments."""
        with tempfile.TemporaryDirectory(prefix="modest") as directory_name:
            output_file = pathlib.Path(directory_name) / "output.json"
            command: Command = (
                self.executable,
                "check",
                "-O",
                output_file,
                "json",
                *map(str, arguments),
            )
            try:
                process = subprocess.run(
                    command,
                    env=self.environment,
                    timeout=timeout,
                    capture_output=capture_output,
                )
            except subprocess.TimeoutExpired as timeout_error:
                raise ToolTimeoutError(
                    "timeout expired during execution of `modest check`",
                    command=command,
                    stdout=timeout_error.stdout,
                    stderr=timeout_error.stderr,
                )
            if process.returncode != 0:
                raise ToolError(
                    f"`modest check` terminated with non-zero returncode {process.returncode}",
                    command=command,
                    stdout=process.stdout,
                    stderr=process.stderr,
                    returncode=process.returncode,
                )
            try:
                return json.loads(output_file.read_text(encoding="utf-8-sig"))
            except FileNotFoundError:
                raise ToolError(
                    "`modest check` did not generate an output file",
                    command=command,
                    stdout=process.stdout,
                    stderr=process.stderr,
                    returncode=0,
                )


@d.dataclass(frozen=True, eq=False)
class ModestChecker(checkers.Checker):
    toolset: Toolset

    @property
    def description(self) -> str:
        return "Modest (mcsta)"

    def check(
        self,
        network: model.Network,
        *,
        properties: t.Optional[checkers.Properties] = None,
        property_names: t.Optional[t.Iterable[str]] = None,
    ) -> checkers.Result:
        with tempfile.TemporaryDirectory(prefix="modest") as directory_name:
            input_file = pathlib.Path(directory_name) / "input.jani"
            named_properties: t.Dict[str, model.Expression] = {}
            if properties is None and property_names is None:
                named_properties.update(
                    {
                        definition.name: definition.expression
                        for definition in network.ctx.properties.values()
                    }
                )
            if properties is not None:
                named_properties.update(properties)
            if property_names is not None:
                for name in property_names:
                    named_properties[
                        name
                    ] = network.ctx.get_property_definition_by_name(name).expression
            input_file.write_text(
                dump_model(network, properties=named_properties), encoding="utf-8"
            )
            result = self.toolset.check((input_file,))
            return {
                dataset["property"]: fractions.Fraction(dataset["value"])
                for dataset in result["data"]
                if "property" in dataset
            }


toolset = Toolset("modest")
checker = ModestChecker(toolset)


def get_checker(*, accept_license: bool) -> checkers.Checker:
    """Returns an instance of :class:`~momba.analysis.Checker`."""
    assert accept_license, "you need to accept the license of the Modest Toolset"
    _setup_locally()
    toolset = Toolset(_MODEST_EXE)
    return ModestChecker(toolset)



# -*- coding:utf-8 -*-
#
# Copyright (C) 2021, Saarland University
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t


@d.dataclass(frozen=True)
class Value:
    """
    Represents a value.
    """

    _value: t.Any

    def __repr__(self) -> str:
        return f"Value({self._value})"

    @property
    def is_int(self) -> bool:
        """
        Returns :code:`True` if the value is an integer.
        """
        return isinstance(self._value, int)

    @property
    def is_float(self) -> bool:
        """
        Returns :code:`True` if the value is a float.
        """
        return isinstance(self._value, float)

    @property
    def is_bool(self) -> bool:
        """
        Returns :code:`True` if the value is a boolean.
        """
        return isinstance(self._value, bool)

    @property
    def is_array(self) -> bool:
        """
        Returns :code:`True` if the value is an array.
        """
        return isinstance(self._value, list)

    @property
    def as_int(self) -> int:
        """
        Asserts that the value is an integer and returns the integer.
        """
        assert self.is_int
        return self._value

    @property
    def as_float(self) -> float:
        """
        Asserts that the value is a float and returns the float.
        """
        assert isinstance(self._value, float)
        return self._value

    @property
    def as_bool(self) -> bool:
        """
        Asserts that the value is a boolean and returns the boolean.
        """
        assert isinstance(self._value, bool)
        return self._value

    @property
    def as_array(self) -> t.Tuple[Value, ...]:
        """
        Asserts that the value is an array and returns a tuple of values.
        """
        assert isinstance(self._value, list)
        return tuple(map(Value, self._value))



# -*- coding:utf-8 -*-
#
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import pathlib

import click

from ..engine import translator

from . import parse


@click.group()
def main() -> None:
    """
    Toolset for working with MOML models.
    """


@main.command()
@click.argument(
    "model_path",
    metavar="MODEL",
    type=click.Path(exists=True, dir_okay=False, readable=True),
)
@click.argument(
    "output_path",
    metavar="OUTPUT",
    type=click.Path(exists=False, dir_okay=False, writable=True),
)
@click.option(
    "--network", "network_name", help="The name of the network to translate.", type=str
)
def translate(model_path: str, output_path: str, network_name: t.Optional[str]) -> None:
    """
    Translates a MOML model to MombaIR.
    """

    ctx = parse(pathlib.Path(model_path).read_text("utf-8"))

    if network_name is None:
        if len(ctx.networks) != 1:
            network_names = {network.name for network in ctx.networks}
            print(
                f"Please specify a network name. Valid network names are {network_names}."
            )
            return
        (network,) = ctx.networks
    else:
        network = ctx.get_network_by_name(network_name)

    translation = translator.translate_network(network)

    pathlib.Path(output_path).write_text(translation.json_network, "utf-8")


if __name__ == "__main__":
    main()



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import functools


T = t.TypeVar("T")


# XXX: this is here just to make `mypy` happy
def lru_cache(func: T) -> T:
    return functools.lru_cache()(func)  # type: ignore



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import enum

from . import types


_JANI_NAME_MAP: t.Dict[str, DistributionType] = {}


class DistributionType(enum.Enum):
    """
    An enum of distribution type.
    """

    DISCRETE_UNIFORM = (
        "DiscreteUniform",
        (
            types.INT,
            types.INT,
        ),
        types.INT,
    )
    """ Discrete uniform distribution. """

    BERNOULLI = "Bernoulli", (types.REAL,)
    """ Bernoulli distribution. """

    BINOMIAL = "Binomial", (types.REAL,)
    """ Binomial distribution. """

    NEGATIVE_BINOMIAL = "NegativeBinomial", (
        types.REAL,
        types.REAL,
    )
    """ Negative binomial distribution. """

    POISSON = "Poisson", (types.REAL,)
    """ Poisson distribution. """

    GEOMETRIC = "Geometric", (types.REAL,)
    """ Geometric distribution. """

    HYPERGEOMETRIC = "Hypergeometric", (
        types.INT,
        types.INT,
        types.INT,
    )
    """ Hypergeometric distribution. """

    CONWAY_MAXWELL_POISSON = "ConwayMaxwellPoisson", (
        types.REAL,
        types.REAL,
    )
    """ Conway Maxwell Poisson distribution. """

    ZIPF = "ZipF", (types.REAL,)
    """ ZipF distribution. """

    UNIFORM = "Uniform", (
        types.REAL,
        types.REAL,
    )
    """ Uniform distribution. """

    NORMAL = "Normal", (
        types.REAL,
        types.REAL,
    )
    """ Normal distribution. """

    LOG_NORMAL = "LogNormal", (
        types.REAL,
        types.REAL,
    )
    """ Logarithmic normal distribution. """

    BETA = "Beta", (
        types.REAL,
        types.REAL,
    )
    """ Beta distribution. """

    CAUCHY = "Cauchy", (
        types.REAL,
        types.REAL,
    )
    """ Cauchy distribution. """

    CHI = "Chi", (types.INT,)
    """ Chi distribution. """

    CHI_SQUARED = "ChiSquared", (types.INT,)
    """ Chi squared distribution. """

    ERLANG = "Erlang", (
        types.INT,
        types.REAL,
    )
    """ Erlang distribution. """

    EXPONENTIAL = "Exponential", (types.REAL,)
    """ Exponential distribution. """

    FISHER_SNEDECOR = "FisherSnedecor", (
        types.REAL,
        types.REAL,
    )
    """ Fisher Snedecor distribution. """

    GAMMA = "Gamma", (
        types.REAL,
        types.REAL,
    )
    """ Gamma distribution. """

    INVERSE_GAMMA = "InverseGamma", (
        types.REAL,
        types.REAL,
    )
    """ Inverse gamma distribution. """

    LAPLACE = "Laplace", (
        types.REAL,
        types.REAL,
    )
    """ Laplace distribution. """

    PARETO = "Pareto", (
        types.REAL,
        types.REAL,
    )
    """ Pareto distribution. """

    RAYLEIGH = "Rayleigh", (types.REAL,)
    """ Rayleigh distribution. """

    STABLE = "Stable", (
        types.REAL,
        types.REAL,
        types.REAL,
        types.REAL,
    )
    """ Stable distribution. """

    STUDENT_T = "StudentT", (
        types.REAL,
        types.REAL,
        types.REAL,
    )
    """ StudentT distribution. """

    WEIBULL = "Weibull", (
        types.REAL,
        types.REAL,
    )
    """ Weibull distribution. """

    TRIANGULAR = "Triangular", (
        types.REAL,
        types.REAL,
        types.REAL,
    )
    """ Triangular distribution. """

    jani_name: str
    parameter_types: t.Tuple[types.Type, ...]
    result_type: types.Type

    def __init__(
        self,
        jani_name: str,
        parameter_types: t.Tuple[types.Type, ...],
        result_type: types.Type = types.REAL,
    ) -> None:
        _JANI_NAME_MAP[jani_name] = self
        self.jani_name = jani_name
        self.parameter_types = parameter_types
        self.result_type = result_type

    @property
    def arity(self) -> int:
        return len(self.parameter_types)

    @staticmethod
    def by_name(jani_name: str) -> DistributionType:
        return _JANI_NAME_MAP[jani_name]



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import enum
import fractions
import math

from . import types


class Operator:
    """
    The base class of all operators.

    Attributes
    ----------
    symbol:
        Symbol associated with the operator.
    """

    name: str
    symbol: str

    def __init__(self, symbol: str):
        self.symbol = symbol


# XXX: This is needed to make operators pickleable. Instead of pickling by value, we pickle
# enums by their name in the enclosing global scope. I am not sure why this is not the
# default behavior as enums are singletons.
#
# Implementing this method directly on `Operator` makes MyPy complain about improper types,
# hence, we add this method retrospectively.
Operator.__reduce_ex__ = (  # type: ignore
    lambda self, protocol: f"{self.__class__.__name__}.{self.name}"  # type: ignore
)


class BinaryOperator(Operator):
    """
    Base class for all binary operators.
    """


class NativeBooleanFunction(t.Protocol):
    def __call__(self, left: bool, right: bool) -> bool:
        pass


class BooleanOperator(BinaryOperator, enum.Enum):
    """
    An enum of boolean operators.
    """

    AND = "∧", lambda left, right: left and right
    """ Logical conjunction. """

    OR = "∨", lambda left, right: left or right
    """ Logical disjunction. """

    # requires JANI extension `derived-operators`
    IMPLY = "⇒", lambda left, right: not left or right
    """ Logical implication. """

    # requires JANI extension `x-momba-operators`
    XOR = "⊕", lambda left, right: (left or right) and not (left and right)
    """ Logical exclusive disjunction. """

    # requires JANI extension `x-momba-operators`
    EQUIV = "⇔", lambda left, right: left is right
    """ Logical equivalence. """

    native_function: NativeBooleanFunction

    def __init__(self, symbol: str, native_function: NativeBooleanFunction) -> None:
        super().__init__(symbol)
        self.native_function = native_function


Number = t.Union[int, float, fractions.Fraction]


class NativeBinaryArithmeticFunction(t.Protocol):
    def __call__(self, left: Number, right: Number) -> Number:
        pass


class ArithmeticBinaryOperator(BinaryOperator, enum.Enum):
    """
    An enum of arithmetic binary operators.
    """

    ADD = "+", lambda left, right: left + right
    """ Addition. """

    SUB = "-", lambda left, right: left - right
    """ Substraction. """

    MUL = "*", lambda left, right: left * right
    """ Multiplication. """

    MOD = "%", lambda left, right: left % right
    """ Euclidean remainder. """

    REAL_DIV = (
        "/",
        lambda left, right: fractions.Fraction(left) / fractions.Fraction(right),
    )
    """ Real devision. """

    LOG = "log", lambda left, right: math.log(left, right)
    """ Logarithm. """

    POW = "pow", lambda left, right: pow(left, right)
    """ Power. """

    # requires JANI extension `derived-operators`
    MIN = "min", lambda left, right: min(left, right)
    """ Minimum. """

    # requires JANI extension `derived-operators`
    MAX = "max", lambda left, right: max(left, right)
    """ Maximum. """

    # requires JANI extension `x-momba-operators`
    FLOOR_DIV = "//", lambda left, right: left // right
    """ Euclidean division. """

    native_function: NativeBinaryArithmeticFunction

    def __init__(
        self, symbol: str, native_function: NativeBinaryArithmeticFunction
    ) -> None:
        super().__init__(symbol)
        self.native_function = native_function


class NativeEqualityFunction(t.Protocol):
    def __call__(self, left: t.Any, right: t.Any) -> bool:
        pass


class EqualityOperator(BinaryOperator, enum.Enum):
    """
    An enum of equality operators.
    """

    EQ = "=", lambda left, right: left == right
    """ Is equal. """

    NEQ = "≠", lambda left, right: left != right
    """ Is not equal. """

    native_function: NativeEqualityFunction

    def __init__(self, symbol: str, native_function: NativeEqualityFunction) -> None:
        super().__init__(symbol)
        self.native_function = native_function


class NativeComparisonFunction(t.Protocol):
    def __call__(self, left: t.Any, right: t.Any) -> bool:
        pass


class ComparisonOperator(BinaryOperator, enum.Enum):
    """
    An enum of comparison operators.
    """

    LT = "<", True, lambda left, right: left < right
    """ Is less than. """

    LE = "≤", False, lambda left, right: left <= right
    """ Is less than or equal to. """

    # requires JANI extension `derived-operators`
    GE = "≥", False, lambda left, right: left >= right
    """ Is greater than or equal to. """

    # requires JANI extension `derived-operators`
    GT = ">", True, lambda left, right: left > right
    """ Is greater than. """

    is_strict: bool
    native_function: NativeComparisonFunction

    def __init__(
        self, symbol: str, is_strict: bool, native_function: NativeComparisonFunction
    ) -> None:
        super().__init__(symbol)
        self.is_strict = is_strict
        self.native_function = native_function

    def swap(self) -> ComparisonOperator:
        return _COMPARISON_SWAP_TABLE[self]

    @property
    def is_less(self) -> bool:
        return self in {ComparisonOperator.LT, ComparisonOperator.LE}

    @property
    def is_greater(self) -> bool:
        return self in {ComparisonOperator.GE, ComparisonOperator.GT}


_COMPARISON_SWAP_TABLE = {
    ComparisonOperator.LT: ComparisonOperator.GT,
    ComparisonOperator.LE: ComparisonOperator.GE,
    ComparisonOperator.GE: ComparisonOperator.LE,
    ComparisonOperator.GT: ComparisonOperator.LT,
}


class UnaryOperator(Operator):
    """
    Base class for all unary operators.
    """


class NotOperator(UnaryOperator, enum.Enum):
    NOT = "¬"
    """ Logical negation. """


class NativeUnaryArithmeticFunction(t.Protocol):
    def __call__(self, operand: Number) -> int:
        pass


class ArithmeticUnaryOperator(UnaryOperator, enum.Enum):
    CEIL = "ceil", lambda _: types.INT, lambda operand: math.ceil(operand)
    """ Round up. """

    FLOOR = "floor", lambda _: types.INT, lambda operand: math.floor(operand)
    """ Round down. """

    # requires JANI extension `derived-operators`
    ABS = "abs", lambda typ: typ, lambda operand: abs(operand)
    """ Absolute value. """

    SGN = (
        "sgn",
        lambda _: types.INT,
        lambda operand: -1 if operand < 0 else (1 if operand > 0 else 0),
    )
    """ Sign. """

    TRC = "trc", lambda _: types.INT, lambda operand: math.trunc(operand)
    """ Truncate. """

    infer_result_type: types.InferTypeUnary
    native_function: NativeUnaryArithmeticFunction

    def __init__(
        self,
        symbol: str,
        infer_result_type: types.InferTypeUnary,
        native_function: NativeUnaryArithmeticFunction,
    ) -> None:
        super().__init__(symbol)
        self.infer_result_type = infer_result_type
        self.native_function = native_function


class MinMax(Operator, enum.Enum):
    """
    Minimum and maximum functions.
    """

    MIN = "min"
    """ Minimum. """

    MAX = "max"
    """ Maximum. """


class Quantifier(Operator, enum.Enum):
    """
    Logical quantifier.
    """

    FORALL = "∀"
    """ Universal quantifier. """

    EXISTS = "∃"
    """ Existential quantifier. """


class BinaryPathOperator(Operator, enum.Enum):
    """
    LTL binary path operators.
    """

    UNTIL = "U"
    """ Until. """

    WEAK_UNTIL = "W"
    """ Weak until. """

    # requires JANI extension `derived-operators`
    RELEASE = "R"
    """ Release. """


class UnaryPathOperator(Operator, enum.Enum):
    """LTL unary path operators."""

    # requires JANI extension `derived-operators`
    EVENTUALLY = "F"
    """ Eventually. """

    GLOBALLY = "G"
    """ Globally. """


class AggregationFunction(Operator, enum.Enum):
    """
    Aggregation functions.
    """

    MIN = "min", {types.REAL}, lambda _: types.REAL
    """ Minimum. """

    MAX = "max", {types.REAL}, lambda _: types.REAL
    """ Maximum. """

    SUM = "sum", {types.REAL}, lambda _: types.REAL
    """ Sum. """

    AVG = "avg", {types.REAL}, lambda _: types.REAL
    """ Average. """

    COUNT = "count", {types.BOOL}, lambda _: types.INT
    """ Count. """

    EXISTS = "∃", {types.BOOL}, lambda _: types.BOOL
    """ Exists. """

    FORALL = "∀", {types.BOOL}, lambda _: types.BOOL
    """ For all. """

    ARGMIN = "argmin", {types.REAL}, lambda _: types.set_of(types.STATE)
    """ Minimizing argument. """

    ARGMAX = "argmax", {types.REAL}, lambda _: types.set_of(types.STATE)
    """ Maximizing argument. """

    VALUES = "values", {types.REAL, types.BOOL}, lambda typ: types.set_of(typ)
    """ Values. """

    allowed_values_type: t.Set[types.Type]
    infer_result_type: types.InferTypeUnary

    def __init__(
        self,
        symbol: str,
        allowed_values_type: t.Set[types.Type],
        infer_result_type: types.InferTypeUnary,
    ) -> None:
        super().__init__(symbol)
        self.allowed_values_type = allowed_values_type
        self.infer_result_type = infer_result_type


class TrigonometricFunction(UnaryOperator, enum.Enum):
    """
    Trigonometric functions.
    """

    SIN = "sin"
    """ Sine. """

    COS = "cos"
    """ Cosine. """

    TAN = "tan"
    """ Tangent. """

    COT = "cot"
    """ Cotangent.. """

    SEC = "sec"
    """ Secant. """

    CSC = "csc"
    """ Cosecant. """

    ARC_SIN = "asin"
    """ Inverse sine """

    ARC_COS = "acos"
    """ Inverse cosine. """

    ARC_TAN = "atan"
    """ Inverse tangent. """

    ARC_COT = "acot"
    """ Inverse cotangent. """

    ARC_SEC = "asec"
    """ Inverse secant. """

    ARC_CSC = "acsc"
    """ Inverse coscant. """



# -*- coding:utf-8 -*-
#
# Copyright (C) 2021, Saarland University
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

import dataclasses as d
import typing as t

from . import actions, expressions


@d.dataclass(frozen=True)
class Observation:
    action_type: actions.ActionType
    arguments: t.Tuple[expressions.Expression, ...]
    probability: t.Optional[expressions.Expression]



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t


def get_subclasses(cls: type, recursive: bool = True) -> t.AbstractSet[type]:
    subclasses: t.Set[type] = set()
    if recursive:
        queue = [cls]
        while queue:
            element = queue.pop()
            subclasses.update(element.__subclasses__())
            queue.extend(element.__subclasses__())
    else:
        subclasses.update(cls.__subclasses__())
    return subclasses



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>
#
# type: ignore

"""
A formal methods based toolbox for reinforcement learning.
"""

from __future__ import annotations

import typing as t

from .. import model, engine

from . import abstract, env, checker, generic


def create_generic_env(
    network: model.Network,
    controlled_instance: model.Instance,
    property_name: str,
    *,
    parameters: engine.Parameters = None,
    rewards: generic.Rewards = generic.DEFAULT_REWARD_STRUCTURE,
    actions: generic.Actions = generic.Actions.EDGE_BY_INDEX,
    observations: generic.Observations = generic.Observations.GLOBAL_ONLY,
    renderer: t.Optional[env.Renderer] = None,
) -> env.MombaEnv:
    """
    Constructs a generic training environment from a JANI model based on the provided options.

    Arguments:
        network: A JANI automaton network.
        controlled_instance: An instance of an automaton in the provided network. The
            decision-making agent trained on the resulting environment is assumed to act
            by resolving the non-determinism in this automaton.
        property_name: The name of a reach-avoid JANI property (specified as part of the
            JANI model the network originates from) for which the agent should be trained.
        parameters: Allows defining values for parameters of the JANI model.
        rewards: Specifies the reward structure used for training.
        actions: Specifies the action space for the environment.
        observations: Specifies the observation space for the environment.
        renderer: Is an optional renderer for the OpenAI Gym API.
    """
    return env.MombaEnv(
        generic.GenericExplorer.create(
            engine.Explorer.new_discrete_time(network, parameters=parameters),
            controlled_instance,
            property_name,
            rewards=rewards,
            actions=actions,
            observations=observations,
        ),
        renderer=renderer,
    )


__all__ = ["abstract", "env", "checker", "create_generic_env"]



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import enum
import random
import warnings

from .. import engine, model
from ..engine import explore

from . import abstract


class Actions(enum.Enum):
    """Specifies how edges of the controlled automaton are chosen."""

    EDGE_BY_INDEX = enum.auto()
    """The edge is chosen based on its index."""

    EDGE_BY_LABEL = enum.auto()
    """The edge is chosen based on its label. """


class Observations(enum.Enum):
    """Specifies what is observable by the agent."""

    GLOBAL_ONLY = enum.auto()
    """Only global variables are observable."""

    LOCAL_AND_GLOBAL = enum.auto()
    """Local and global variables are observable."""

    OMNISCIENT = enum.auto()
    """All (non-transient) variables are observable."""


class _ActionResolver:
    num_actions: int

    def available(self, state: engine.State[engine.DiscreteTime]) -> t.Sequence[bool]:
        raise NotImplementedError()

    def resolve(
        self,
        transitions: t.Sequence[engine.Transition[engine.DiscreteTime]],
        action: int,
    ) -> t.Sequence[engine.Transition[engine.DiscreteTime]]:
        raise NotImplementedError()


@d.dataclass(frozen=True)
class _EdgeByIndexResolver(_ActionResolver):
    instance: model.Instance
    num_actions: int

    @classmethod
    def create(cls, instance: model.Instance) -> _EdgeByIndexResolver:
        return cls(instance, len(instance.automaton.edges))

    def available(self, state: engine.State[engine.DiscreteTime]) -> t.Sequence[bool]:
        available_actions: t.Set[int] = set()
        for transition in state.transitions:
            available_actions.add(transition.index_vector.get(self.instance, -1))
        available_actions.discard(-1)
        return [action in available_actions for action in range(self.num_actions)]

    def resolve(
        self,
        transitions: t.Sequence[engine.Transition[engine.DiscreteTime]],
        action: int,
    ) -> t.Sequence[engine.Transition[engine.DiscreteTime]]:
        return [
            transition
            for transition in transitions
            if transition.index_vector.get(self.instance, -1) == action
        ]


@d.dataclass(frozen=True)
class _EdgeByLabelResolver(_ActionResolver):
    instance: model.Instance
    num_actions: int
    action_mapping: t.Mapping[int, model.ActionType] = d.field(default_factory=dict)
    reverse_action_mapping: t.Mapping[model.ActionType, int] = d.field(
        default_factory=dict
    )

    @classmethod
    def create(
        cls, ctx: model.Context, instance: model.Instance
    ) -> _EdgeByLabelResolver:
        action_types: t.Set[model.ActionType] = set()
        for edge in instance.automaton.edges:
            if edge.action_pattern is not None:
                action_types.add(edge.action_pattern.action_type)
        num_actions = len(action_types)
        action_mapping: t.Dict[int, model.ActionType] = {}
        # `ctx.action_types` should be ordered as in the JANI file
        for action_type in ctx.action_types.values():
            if action_type in action_types:
                action_mapping[len(action_mapping)] = action_type
        return cls(
            instance,
            num_actions,
            action_mapping,
            reverse_action_mapping={
                typ: number for number, typ in action_mapping.items()
            },
        )

    def available(self, state: engine.State[engine.DiscreteTime]) -> t.Sequence[bool]:
        available_actions: t.Set[int] = set()
        for transition in state.transitions:
            instance_action = transition.action_vector.get(self.instance, None)
            if instance_action is not None:
                available_actions.add(
                    self.reverse_action_mapping[instance_action.action_type]
                )
            available_actions.add(transition.index_vector.get(self.instance, -1))
        available_actions.discard(-1)
        return [action in available_actions for action in range(self.num_actions)]

    def resolve(
        self,
        transitions: t.Sequence[engine.Transition[engine.DiscreteTime]],
        action: int,
    ) -> t.Sequence[engine.Transition[engine.DiscreteTime]]:
        action_type = self.action_mapping[action]
        result = []
        for transition in transitions:
            instance_action = transition.action_vector.get(self.instance, None)
            if (
                instance_action is not None
                and instance_action.action_type == action_type
            ):
                result.append(transition)
        return transitions


def _create_action_resolver(
    actions: Actions, ctx: model.Context, instance: model.Instance
) -> _ActionResolver:
    if actions is Actions.EDGE_BY_INDEX:
        return _EdgeByIndexResolver.create(instance)
    else:
        assert actions is Actions.EDGE_BY_LABEL
        return _EdgeByLabelResolver.create(ctx, instance)


def _count_features(value: engine.Value) -> int:
    if value.is_int or value.is_bool or value.is_float:
        return 1
    else:
        return sum(map(_count_features, value.as_array))


def _extend_state_vector(vector: t.List[float], value: engine.Value) -> None:
    if value.is_array:
        for value in value.as_array:
            _extend_state_vector(vector, value)
    else:
        vector.append(float(value._value))


@d.dataclass(frozen=True)
class Rewards:
    """Specifies the rewards for a reachability objective."""

    goal_reached: float = 100
    """The reward when a goal state is reached. """

    dead_end: float = -100
    """The reward when a dead end or bad state is reached."""

    step_taken: float = 0
    """The reward when a valid decision has been taken."""

    invalid_action: float = -100
    """The reward when an invalid decision has been taken."""


DEFAULT_REWARD_STRUCTURE = Rewards()


@d.dataclass(frozen=True)
class Objective:
    r"""A reach-avoid objective.

    Represents a reach-avoid objective of the form
    :math:`\lnot\phi\mathbin{\mathbf{U}}\psi`, i.e., :math:`\lnot\phi` has to
    be true until the goal :math:`\psi` is reached. Used in
    conjunction with :class:`Rewards` to provide rewards.
    """

    goal_predicate: model.Expression
    r"""A boolean expression for :math:`\psi`."""

    dead_predicate: model.Expression
    r"""A boolean expression for :math:`\phi`."""


def _extract_property(prop: model.Expression) -> Objective:
    if isinstance(prop, model.properties.Aggregate):
        assert prop.function in {
            model.operators.AggregationFunction.MIN,
            model.operators.AggregationFunction.MAX,
            model.operators.AggregationFunction.VALUES,
            model.operators.AggregationFunction.EXISTS,
            model.operators.AggregationFunction.FORALL,
        }, f"Unsupported aggregation function {prop.function}"
        assert isinstance(
            prop.predicate, model.properties.StateSelector
        ), f"Unsupported state predicate {prop.predicate} in aggregation"
        assert (
            prop.predicate.predicate is model.properties.StatePredicate.INITIAL
        ), "Unsupported state predicate for aggregation."
        prop = prop.values
    if isinstance(prop, model.properties.Probability):
        prop = prop.formula
    if isinstance(prop, model.properties.UnaryPathFormula):
        assert (
            prop.operator is model.operators.UnaryPathOperator.EVENTUALLY
        ), "Unsupported unary path formula."
        return Objective(
            goal_predicate=prop.formula, dead_predicate=model.ensure_expr(False)
        )
    elif isinstance(prop, model.properties.BinaryPathFormula):
        assert (
            prop.operator is model.operators.BinaryPathOperator.UNTIL
        ), "Unsupported binary path formula."
        left = prop.left
        right = prop.right
        return Objective(
            goal_predicate=right, dead_predicate=model.expressions.logic_not(left)
        )
    else:
        raise Exception("Unsupported property!")


@d.dataclass(frozen=True)
class _Context:
    explorer: engine.Explorer[engine.DiscreteTime]

    controlled_instance: model.Instance

    initial_state: engine.State[engine.DiscreteTime]

    objective: Objective

    dead_predicate: explore.CompiledGlobalExpression[engine.DiscreteTime]
    goal_predicate: explore.CompiledGlobalExpression[engine.DiscreteTime]

    rewards: Rewards

    actions: Actions
    observations: Observations

    action_resolver: _ActionResolver

    global_variables: t.Tuple[str, ...]
    local_variables: t.Tuple[str, ...]
    other_variables: t.Mapping[model.Instance, t.Tuple[str, ...]]

    num_features: int

    @classmethod
    def create(
        cls,
        explorer: engine.Explorer[engine.DiscreteTime],
        controlled_instance: model.Instance,
        property_name: str,
        *,
        rewards: t.Optional[Rewards] = None,
        actions: Actions = Actions.EDGE_BY_INDEX,
        observations: Observations = Observations.GLOBAL_ONLY,
    ) -> _Context:
        network = explorer.network
        initial_states = explorer.initial_states
        objective = _extract_property(
            network.ctx.get_property_definition_by_name(property_name).expression
        )
        goal_predicate = explorer.compile_global_expression(objective.goal_predicate)
        dead_predicate = explorer.compile_global_expression(objective.dead_predicate)
        if len(initial_states) != 1:
            raise Exception("Invalid number of initial states.")
        (initial_state,) = initial_states

        num_features = 0

        global_variables = []
        for declaration in network.ctx.global_scope.variable_declarations:
            if declaration.is_transient:
                continue
            global_variables.append(declaration.identifier)
            value = initial_state.global_env[declaration.identifier]
            num_features += _count_features(value)

        local_variables = []
        if observations in {Observations.LOCAL_AND_GLOBAL, Observations.OMNISCIENT}:
            for (
                declaration
            ) in controlled_instance.automaton.scope.variable_declarations:
                if declaration.is_transient:
                    continue
                local_variables.append(declaration.identifier)
                value = initial_state.get_local_env(controlled_instance)[
                    declaration.identifier
                ]
                num_features += _count_features(value)

        other_variables: t.Dict[model.Instance, t.Tuple[str, ...]] = {}
        if observations is Observations.OMNISCIENT:
            for instance in sorted(
                network.instances, key=lambda instance: str(instance.automaton.name)
            ):
                if instance is controlled_instance:
                    # do not include the local variables twice
                    continue
                instance_variables: t.List[str] = []
                for declaration in instance.automaton.scope.variable_declarations:
                    if declaration.is_transient:
                        continue
                    instance_variables.append(declaration.identifier)
                    value = initial_state.get_local_env(instance)[
                        declaration.identifier
                    ]
                    num_features += _count_features(value)
                if instance_variables:
                    other_variables[instance] = tuple(sorted(instance_variables))

        return cls(
            explorer=explorer,
            controlled_instance=controlled_instance,
            initial_state=initial_state,
            objective=objective,
            goal_predicate=goal_predicate,
            dead_predicate=dead_predicate,
            actions=actions,
            observations=observations,
            global_variables=tuple(sorted(global_variables)),
            local_variables=tuple(sorted(local_variables)),
            other_variables=other_variables,
            rewards=rewards or Rewards(),
            action_resolver=_create_action_resolver(
                actions, network.ctx, controlled_instance
            ),
            num_features=num_features,
        )


class GenericExplorer(abstract.Explorer):
    _ctx: _Context

    state: engine.State[engine.DiscreteTime]

    def __init__(
        self,
        *,
        _ctx: _Context,
        _state: t.Optional[engine.State[engine.DiscreteTime]] = None,
    ) -> None:
        self._ctx = _ctx
        self.state = _state or _ctx.initial_state

    @classmethod
    def create(
        cls,
        explorer: engine.Explorer[engine.DiscreteTime],
        controlled_instance: model.Instance,
        property_name: str,
        *,
        rewards: t.Optional[Rewards] = None,
        actions: Actions = Actions.EDGE_BY_INDEX,
        observations: Observations = Observations.GLOBAL_ONLY,
    ) -> GenericExplorer:
        ctx = _Context.create(
            explorer,
            controlled_instance,
            property_name,
            rewards=rewards,
            actions=actions,
            observations=observations,
        )
        return cls(_ctx=ctx, _state=ctx.initial_state)

    def _has_choice(self, state: engine.State[engine.DiscreteTime]) -> bool:
        return any(
            self._ctx.controlled_instance in transition.index_vector
            for transition in state.transitions
        )

    def _is_final(self, state: engine.State[engine.DiscreteTime]) -> bool:
        return self._has_choice(state) or self._has_terminated(state)

    def _explore_until_choice(self) -> None:
        while not self._is_final(self.state):
            if len(self.state.transitions) > 1:
                warnings.warn(
                    "Uncontrolled nondeterminism has been resolved uniformly."
                )
            self.state = random.choice(self.state.transitions).destinations.pick().state

    def _explore_successors(
        self, state: engine.State[engine.DiscreteTime]
    ) -> t.Dict[engine.State[engine.DiscreteTime], float]:
        pending = [(1.0, state)]
        result: t.Dict[engine.State[engine.DiscreteTime], float] = {}
        while pending:
            probability, state = pending.pop()
            if self._is_final(state):
                if state not in result:
                    result[state] = 0.0
                result[state] += probability
            else:
                if len(state.transitions) > 1:
                    warnings.warn(
                        "Uncontrolled nondeterminism has been resolved uniformly."
                    )
                transition_probability = probability / len(state.transitions)
                for transition in state.transitions:
                    for destination in transition.destinations.support:
                        pending.append(
                            (
                                transition_probability
                                * float(
                                    transition.destinations.get_probability(destination)
                                ),
                                destination.state,
                            )
                        )
        return result

    @property
    def num_actions(self) -> int:
        return self._ctx.action_resolver.num_actions

    @property
    def num_features(self) -> int:
        return self._ctx.num_features

    def _state_vector(
        self, state: engine.State[engine.DiscreteTime]
    ) -> t.Sequence[float]:
        vector: t.List[float] = []
        for variable in self._ctx.global_variables:
            value = state.global_env[variable]
            _extend_state_vector(vector, value)
        local_env = state.get_local_env(self._ctx.controlled_instance)
        for variable in self._ctx.local_variables:
            _extend_state_vector(vector, local_env[variable])
        for instance, variables in self._ctx.other_variables.items():
            local_env = state.get_local_env(instance)
            for variable in variables:
                _extend_state_vector(vector, local_env[variable])
        return tuple(vector)

    @property
    def state_vector(self) -> t.Sequence[float]:
        return self._state_vector(self.state)

    def _has_terminated(self, state: engine.State[engine.DiscreteTime]) -> bool:
        is_dead = self._ctx.dead_predicate.evaluate(state).as_bool
        return not state.transitions or is_dead or self._has_reached_goal(state)

    def _has_reached_goal(self, state: engine.State[engine.DiscreteTime]) -> bool:
        return self._ctx.goal_predicate.evaluate(state).as_bool

    def _get_reward(self, state: engine.State[engine.DiscreteTime]) -> float:
        if self._has_reached_goal(state):
            return self._ctx.rewards.goal_reached
        elif self._has_terminated(state):
            return self._ctx.rewards.dead_end
        else:
            return self._ctx.rewards.step_taken

    @property
    def has_terminated(self) -> bool:
        return self._has_terminated(self.state)

    @property
    def has_reached_goal(self) -> bool:
        return self._has_reached_goal(self.state)

    @property
    def available_actions(self) -> t.Sequence[bool]:
        return tuple(self._ctx.action_resolver.available(self.state))

    @property
    def available_transitions(self) -> t.Sequence[abstract.Transition]:
        if self.has_terminated:
            return []
        result = []
        for action, available in enumerate(self.available_actions):
            if not available:
                continue
            transitions = self._ctx.action_resolver.resolve(
                self.state.transitions, action
            )
            if len(transitions) > 1:
                warnings.warn(
                    "Uncontrolled nondeterminism has been resolved uniformly."
                )
            transition_probability = 1.0 / len(transitions)
            destinations: t.Dict[t.Sequence[float], abstract.Destination] = {}
            for transition in transitions:
                for destination in transition.destinations.support:
                    for (
                        final_successor,
                        probability,
                    ) in self._explore_successors(destination.state).items():
                        final_successor_vector = self._state_vector(final_successor)
                        reward = self._get_reward(final_successor)
                        probability *= transition_probability
                        probability *= transition.destinations.get_probability(
                            destination
                        )
                        if final_successor_vector in destinations:
                            probability += destinations[
                                final_successor_vector
                            ].probability
                        destinations[final_successor_vector] = abstract.Destination(
                            final_successor_vector, reward, probability
                        )
            result.append(
                abstract.Transition(action, destinations=tuple(destinations.values()))
            )
        return result

    def step(self, action: int) -> float:
        if self.has_terminated:
            if self.has_reached_goal:
                return self._ctx.rewards.goal_reached
            else:
                return self._ctx.rewards.dead_end
        selected_transitions = self._ctx.action_resolver.resolve(
            self.state.transitions, action
        )
        if not selected_transitions:
            return self._ctx.rewards.invalid_action
        else:
            if len(selected_transitions) > 1:
                warnings.warn(
                    "Uncontrolled nondeterminism has been resolved uniformly."
                )
            self.state = random.choice(selected_transitions).destinations.pick().state
            self._explore_until_choice()
        return self._get_reward(self.state)

    def reset(self, *, explore_until_choice: bool = True) -> None:
        self.state = self._ctx.initial_state
        if explore_until_choice:
            self._explore_until_choice()

    def fork(self) -> GenericExplorer:
        return GenericExplorer(_ctx=self._ctx, _state=self.state)



# -*- coding:utf-8 -*-
#
# Copyright (C) 2021, Saarland University
# Copyright (C) 2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import json
import functools

from .. import model
from ..model import expressions
from ..utils.distribution import Distribution

from .time import DiscreteTime, TimeType, CompiledNetwork
from .values import Value


TimeTypeT = t.TypeVar("TimeTypeT", bound=TimeType)

Parameters = t.Optional[t.Mapping[str, expressions.ValueOrExpression]]

_cache_explored_states = True


def disable_exploration_cache() -> None:
    global _cache_explored_states
    _cache_explored_states = False


@d.dataclass(frozen=True)
class Action:
    """
    Represents an action.

    The *arguments* of an action are usually empty. They are used for
    the experimental value-passing feature of Momba which has not been
    stabilized yet.

    Attributes
    ----------
    action_type:
        The :class:`~momba.model.ActionType` of the action.
    arguments:
        The arguments of the action (a tuple of values).
    """

    action_type: model.ActionType
    arguments: t.Tuple[Value, ...]


@d.dataclass(frozen=True, repr=False)
class Destination(t.Generic[TimeTypeT]):
    """
    Represents a destination of a transition.

    Attributes
    ----------
    explorer:
        The :class:`Explorer` associated with the destination.
    """

    explorer: Explorer[TimeTypeT]

    _state: t.Any
    _transition: t.Any
    _destination: t.Any

    @property
    def probability(self) -> float:
        """
        The probability associated with the destination.
        """
        return self._destination.probability()

    @property
    def state(self) -> State[TimeTypeT]:
        """
        The target :class:`State` associated with the destination.
        """
        cached_state = self.__dict__.get("_cached_state", None)
        if cached_state is None:
            cached_state = State(
                self.explorer,
                self._destination.successor(),
            )
            if _cache_explored_states:
                self.__dict__["_cached_state"] = cached_state
        return cached_state


def _action(action: t.Any, explorer: Explorer[TimeTypeT]) -> t.Optional[Action]:
    if action.is_silent():
        return None
    label = action.label()
    assert isinstance(label, str)
    arguments = action.arguments()
    return Action(
        explorer.network.ctx.get_action_type_by_name(label),
        tuple(Value(value) for value in arguments),
    )


@d.dataclass(frozen=True, repr=False)
class Transition(t.Generic[TimeTypeT]):
    """
    Represents a joint transition of an automaton network.

    Attributes
    ----------
    explorer:
        The :class:`Explorer` associated with the transition.
    instances:
        The automaton instances participating in the transition.
    action:
        The action associated with the transition.
    action_vector:
        The actions with which the respective instances participate.

        Is a mapping from instances to actions.
    edge_vector:
        The edges with which the respective instances participate.

        Is a mapping from instances to edges.
    destinations:
        The probability distribution over destinations.
    """

    explorer: Explorer[TimeTypeT]
    source: State[TimeTypeT]

    _state: t.Any
    _transition: t.Any

    @property
    def instances(self) -> t.AbstractSet[model.Instance]:
        instances = set()
        for edge_reference in json.loads(self._transition.edge_vector()):
            automaton_name = edge_reference["location"]["automaton"]["name"]
            instance = self.explorer._compiled.translation.instance_name_to_instance[
                automaton_name
            ]
            instances.add(instance)
        return instances

    @property
    def action(self) -> t.Optional[Action]:
        return _action(self._transition.action(), self.explorer)

    @property
    def action_vector(self) -> t.Mapping[model.Instance, t.Optional[Action]]:
        action_vector = {}
        for edge_reference, action in zip(
            json.loads(self._transition.edge_vector()), self._transition.action_vector()
        ):
            automaton_name = edge_reference["location"]["automaton"]["name"]
            instance = self.explorer._compiled.translation.instance_name_to_instance[
                automaton_name
            ]
            action_vector[instance] = _action(action, self.explorer)
        return action_vector

    @property
    def edge_vector(self) -> t.Mapping[model.Instance, model.Edge]:
        return {
            instance: instance.automaton.edges[index]
            for instance, index in self.index_vector.items()
        }

    @functools.cached_property
    def index_vector(self) -> t.Mapping[model.Instance, int]:
        return {
            self.explorer._compiled.translation.instance_vector[
                instance_index
            ]: edge_index
            for instance_index, edge_index in self._transition.numeric_reference_vector()
        }

    @functools.cached_property
    def destinations(self) -> Distribution[Destination[TimeTypeT]]:
        destinations = tuple(
            Destination(self.explorer, self._state, self._transition, destination)
            for destination in self._transition.destinations()
        )
        return Distribution(
            {destination: destination.probability for destination in destinations}
        )

    @functools.cached_property
    def valuations(self) -> TimeTypeT:
        return self.explorer.time_type.load_valuations(self._transition.valuations())


@d.dataclass(frozen=True, repr=False)
class State(t.Generic[TimeTypeT]):
    """
    Represents a state of an automaton network.

    Attributes
    ----------
    explorer:
        The :class:`Explorer` associated with the state.
    """

    explorer: Explorer[TimeTypeT]

    _state: t.Any

    @functools.cached_property
    def global_env(self) -> t.Mapping[str, Value]:
        """
        The global environment, i.e., a mapping from global variables to values.
        """
        declarations = self.explorer._compiled.translation.declarations
        return {
            name: Value(self._state.get_global_value(declaration.identifier))
            for name, declaration in declarations.globals_table.items()
            if not declaration.is_transient
        }

    def get_local_env(self, instance: model.Instance) -> t.Mapping[str, Value]:
        """
        Returns the local environment of the provided automaton instance.
        """
        declarations = self.explorer._compiled.translation.declarations
        return {
            name: Value(self._state.get_global_value(declaration.identifier))
            for name, declaration in declarations.locals_table[instance].items()
        }

    @functools.cached_property
    def locations(self) -> t.Mapping[model.Instance, model.Location]:
        """
        The locations of the respective automata instances.

        A mapping from instances to locations.
        """
        return {
            instance: self.explorer._compiled.translation.reversed_instance_to_location_names[
                instance
            ][
                self._state.get_location_of(name)
            ]
            for instance, name in self.explorer._compiled.translation.instance_names.items()
        }

    @functools.cached_property
    def transitions(self) -> t.Sequence[Transition[TimeTypeT]]:
        """
        Outgoing transitions of the state.
        """
        return tuple(
            Transition(self.explorer, self, self._state, transition)
            for transition in self._state.transitions()
        )

    @property
    def valuations(self) -> TimeTypeT:
        return self.explorer.time_type.load_valuations(self._state.valuations())


@d.dataclass(frozen=True, repr=False)
class CompiledGlobalExpression(t.Generic[TimeTypeT]):
    _compiled: t.Any

    def evaluate(self, state: State[TimeTypeT]) -> Value:
        return Value(_value=self._compiled.evaluate(state._state))


class Explorer(t.Generic[TimeTypeT]):
    """
    Main interface to the state space exploration engine.

    .. warning::
        A network must not be modified once an explorer has been created
        for it.
        Modifying the network nonetheless may lead to all kinds of
        unspecified behavior.

    Paramaters
    ----------

    Attributes
    ----------
    network:
        The :class:`~momba.model.Network` the explorer has been created for.
    time_type:
        The :class:`TimeType` of the explorer.
    """

    network: model.Network
    time_type: t.Type[TimeTypeT]

    _compiled: CompiledNetwork

    def __init__(
        self,
        network: model.Network,
        time_type: t.Type[TimeTypeT],
        *,
        parameters: Parameters = None,
    ) -> None:
        self.network = network
        self.time_type = time_type
        self._compiled = self.time_type.compile(network, parameters=parameters)

    @staticmethod
    def new_discrete_time(
        network: model.Network,
        *,
        parameters: Parameters = None,
    ) -> Explorer[DiscreteTime]:
        """
        Creates a new discrete time explorer.
        """
        return Explorer(network, DiscreteTime, parameters=parameters)

    @functools.cached_property
    def initial_states(self) -> t.AbstractSet[State[TimeTypeT]]:
        """
        The initial states of the network.
        """
        return frozenset(
            State(self, state) for state in self._compiled.internal.initial_states()
        )

    @functools.cached_property
    def _states_and_transitions(self) -> t.Tuple[int, int]:
        return self._compiled.internal.count_states_and_transitions()

    def compile_global_expression(
        self, expr: model.Expression
    ) -> CompiledGlobalExpression[TimeTypeT]:
        json_representation = self._compiled.translation.translate_global_expression(
            expr
        )
        compiled = self._compiled.internal.compile_global_expression(
            json_representation
        )
        return CompiledGlobalExpression(_compiled=compiled)

    def count_states(self) -> int:
        return self._states_and_transitions[0]

    def count_transitions(self) -> int:
        return self._states_and_transitions[1]



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020-2021, Saarland University
# Copyright (C) 2020-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

import typing as t

import collections
import gc
import pathlib
import sys
import time

from mxu.maps import FrozenMap


Classes = t.Union[type, t.Tuple[type, ...]]
Objects = t.Mapping[type, t.Sequence[t.Any]]

ALL_CLASSES: t.Tuple[type, ...] = (object,)

MAX_DEPTH = 5

_CONTAINERS = (list, tuple, dict, FrozenMap)


def fmt_size(
    size: t.Union[float, int],
) -> str:
    result: str
    for unit in ("B", "KB", "MB", "GB"):
        result = f"{size} {unit}"
        if size > 1000:
            size /= 1000
        else:
            break
    return result


def get_objects(classes: Classes = ALL_CLASSES, prefix: str = "momba") -> Objects:
    objects: t.Dict[type, t.List[t.Any]] = collections.defaultdict(list)
    for obj in gc.get_objects():
        cls = obj.__class__
        if cls.__module__.startswith(prefix):
            if issubclass(cls, classes):
                objects[obj.__class__].append(obj)
    return objects


def compute_usage(objects: Objects) -> t.Mapping[type, int]:
    return {
        cls: sum(sys.getsizeof(obj) for obj in cls_objects)
        for cls, cls_objects in objects.items()
    }


def compute_counts(objects: Objects) -> t.Mapping[type, int]:
    return {cls: len(cls_objects) for cls, cls_objects in objects.items()}


def _collect_referrers(obj: t.Any, counts: t.Dict[type, int], depth: int = 0) -> None:
    for referrer in gc.get_referrers(obj):
        if depth < MAX_DEPTH and isinstance(referrer, _CONTAINERS):
            _collect_referrers(referrer, counts, depth + 1)
        else:
            counts[referrer.__class__] += 1


def compute_referrers(objects: Objects) -> t.Mapping[type, t.Mapping[type, int]]:
    referrers: t.Dict[type, t.Dict[type, int]] = {}
    for cls, cls_objects in objects.items():
        counts = referrers[cls] = collections.defaultdict(int)
        for obj in cls_objects:
            _collect_referrers(obj, counts)
    return referrers


def dump_statistics(path: pathlib.Path, prefix: str = "momba") -> None:
    with path.open("at", encoding="utf-8") as file:
        file.write(f"dump_statistics @ {time.monotonic_ns()}\n")
        objects = get_objects(prefix=prefix)
        usage = compute_usage(objects)
        counts = compute_counts(objects)
        referrers = compute_referrers(objects)
        for cls in sorted(objects.keys(), key=lambda cls: usage[cls], reverse=True):
            cls_name = f"{cls.__module__}.{cls.__name__}"
            file.write(
                f"  {cls_name} | count: {counts[cls]} | size: {fmt_size(usage[cls])}\n"
                f"  referrrers: {sum(referrers[cls].values())}\n"
            )
            for ref_cls, ref_count in sorted(
                referrers[cls].items(), key=lambda item: item[1], reverse=True
            ):
                ref_cls_name = f"{ref_cls.__module__}.{ref_cls.__name__}"
                file.write(f"    {ref_cls_name}: {ref_count}\n")



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations


class ModelingError(Exception):
    """
    A general modeling error.
    """


class InvalidTypeError(ModelingError):
    pass


class IncompatibleAssignmentsError(ModelingError):
    pass


class TypeConstructionError(ModelingError):
    pass


class InvalidDeclarationError(ModelingError):
    pass


class UnboundIdentifierError(ModelingError):
    pass


class NotFoundError(ModelingError):
    pass



# -*- coding:utf-8 -*-
#
# Copyright (C) 2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import fractions
import random

from mxu.maps import FrozenMap


ElementT = t.TypeVar("ElementT", bound=t.Hashable)


class Distribution(t.Generic[ElementT]):
    """A probability distribution."""

    _mapping: FrozenMap[ElementT, fractions.Fraction]

    @classmethod
    def create_dirac(cls, element: ElementT) -> Distribution[ElementT]:
        return cls({element: 1})

    @classmethod
    def create_uniform(cls, *elements: ElementT) -> Distribution[ElementT]:
        probability = fractions.Fraction(1, len(elements))
        return cls({element: probability for element in elements})

    def __init__(
        self, mapping: t.Mapping[ElementT, t.Union[int, float, fractions.Fraction]]
    ) -> None:
        self._mapping = FrozenMap.transfer_ownership(
            {
                element: fractions.Fraction(probability)
                for element, probability in mapping.items()
            }
        )
        assert all(probability >= 0 for probability in self._mapping.values())
        # assert sum(self._mapping.values()) == 1

    def __str__(self) -> str:
        return f"Distribution({self._mapping})"

    @property
    def support(self) -> t.List[ElementT]:
        return list(
            element for element, probability in self._mapping.items() if probability > 0
        )

    @property
    def is_dirac(self) -> bool:
        return len(self.support) == 1

    def get_probability(self, element: ElementT) -> fractions.Fraction:
        return self._mapping.get(element, fractions.Fraction(0))

    def pick(self) -> ElementT:
        """Picks an element at random according to the distribution."""
        max_denominator = max(
            probability.denominator for probability in self._mapping.values()
        )
        outcome = fractions.Fraction(
            random.randint(0, max_denominator), max_denominator
        )
        total = fractions.Fraction(0)
        for element in self.support:
            total += self.get_probability(element)
            if outcome <= total:
                return element
        raise RuntimeError("empty distribution is not possible")



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

"""
A collection of data-structures and algorithms useful in the context of quantitative models.
"""

from __future__ import annotations

from . import dbm

from .dbm import DBM
from .intervals import Interval


__all__ = ["dbm", "DBM", "Interval"]



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2021, Saarland University
# Copyright (C) 2019-2021, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import dataclasses as d
import typing as t

import collections
import enum

from . import actions, errors, functions, expressions, types

from .automata import Automaton
from .networks import Network


class ModelType(enum.Enum):
    """
    An enum representing different *model types*.

    Attributes
    ----------
    full_name: str
        The full human-readable name of the model type.
    """

    LTS = "Labeled Transition System"
    """ Labeled Transition System """

    DTMC = "Discrete-Time Markov Chain"
    """ Discrete-Time Markov Chain """

    CTMC = "Continuous-Time Markov Chain"
    """ Continuous-Time Markov Chain """

    MDP = "Markov Decision Process"
    """ Markov Decision Process """

    CTMDP = "Continuous-Time Markov Decision Process"
    """ Continuous-Time Markov Decision Process """

    MA = "Markov Automaton"
    """ Markov Automaton """

    TA = "Timed Automaton"
    """ Timed Automaton """

    PTA = "Probabilistic Timed Automaton"
    """ Probabilistic Timed Automaton """

    STA = "Stochastic Timed Automaton"
    """ Stochastic Timed Automaton """

    HA = "Hybrid Automaton"
    """ Hybrid Automaton """

    PHA = "Probabilistic Hybrid Automaton"
    """ Probabilistic Hybrid Automaton """

    SHA = "Stochastic Hybrid Automaton"
    """ Stochastic Hybrid Automaton """

    full_name: str

    def __init__(self, full_name: str):
        self.full_name = full_name

    @property
    def uses_clocks(self) -> bool:
        """
        Returns :obj:`True` if and only if the respective models use real-value clocks.
        """
        return self in _CLOCK_TYPES

    @property
    def is_untimed(self) -> bool:
        """
        Returns :obj:`True` if and only if the model type is *not timed*.

        Untimed model types are :code:`LTS`, :code:`DTMC`, and :code:`MDP`.
        """
        return self in _UNTIMED_TYPES


_CLOCK_TYPES = {
    ModelType.TA,
    ModelType.PTA,
    ModelType.STA,
    ModelType.HA,
    ModelType.PHA,
    ModelType.SHA,
}

_UNTIMED_TYPES = {ModelType.MDP, ModelType.LTS, ModelType.DTMC}


# XXX: this class should be abstract, however, then it would not type-check
# https://github.com/python/mypy/issues/5374
@d.dataclass(frozen=True)
class IdentifierDeclaration:
    """
    Represents a declaration of an identifier.

    Attributes
    ----------
    identifier:
        The declared identifier.
    typ:
        The type of the identifier.
    comment:
        An additional optional comment for the declaration.
    """

    identifier: str
    typ: types.Type

    comment: t.Optional[str] = None

    # XXX: this method shall be implemented by all subclasses
    def validate(self, scope: Scope) -> None:
        """
        Validates that the declaration is valid in the given scope.

        Raises :class:`~errors.ModelingError` if the declaration is invalid.
        """
        raise NotImplementedError()


@d.dataclass(frozen=True)
class VariableDeclaration(IdentifierDeclaration):
    """
    Represents a *variable declaration*.

    Attributes
    ----------
    is_transient:
        Optional boolean flag indicating whether the variable is *transient*.
    initial_value:
        Optional :class:`~momba.model.Expression` providing an initial value for the variable.
    """

    is_transient: t.Optional[bool] = None
    initial_value: t.Optional[expressions.Expression] = None

    def __post_init__(self) -> None:
        # if self.is_transient and self.initial_value is None:
        #     raise errors.ModelingError(
        #         "transient variables must have an initial value", self
        #     )
        pass

    def validate(self, scope: Scope) -> None:
        if self.initial_value is not None:
            value_type = scope.get_type(self.initial_value)
            if not self.typ.is_assignable_from(value_type):
                raise errors.ModelingError(
                    f"type of initial value {value_type} is not "
                    f"assignable to variable type {self.typ}",
                    self,
                )


@d.dataclass(frozen=True)
class ConstantDeclaration(IdentifierDeclaration):
    """
    Represents a *constant declaration*.

    Attributes
    ----------
    value:
        Optional :class:`~momba.model.Expression` specifying the value of the constant.
    """

    value: t.Optional[expressions.Expression] = None

    @property
    def is_parameter(self) -> bool:
        """
        Returns :obj:`True` if and only if the constant is a *parameter*.

        Parameters are constants without a :attr:`value`.
        """
        return self.value is None

    def validate(self, scope: Scope) -> None:
        if self.value is not None:
            value_type = scope.get_type(self.value)
            if not self.typ.is_assignable_from(value_type):
                raise errors.ModelingError(
                    f"type of constant value {value_type} is not "
                    f"assignable to constant type {self.typ}"
                )


@d.dataclass(frozen=True)
class PropertyDefinition:
    """
    Represents a *property definition*.

    Attributes
    ----------
    name:
        The name of the property.
    expression:
        An :class:`~momba.model.Expression` defining the property.
    comment:
        An optional comment describing the property.
    """

    name: str
    expression: expressions.Expression

    comment: t.Optional[str] = None


class Scope:
    """
    Represents a *scope*.

    Attributes
    ----------
    ctx:
        The modeling context associated with the scope.
    parent:
        The parent scope if it exists (:obj:`None` if there is no parent).
    """

    ctx: Context
    parent: t.Optional[Scope]

    _declarations: t.OrderedDict[str, IdentifierDeclaration]
    _functions: t.OrderedDict[str, functions.FunctionDefinition]

    _type_cache: t.Dict[expressions.Expression, types.Type]

    def __init__(self, ctx: Context, parent: t.Optional[Scope] = None):
        self.ctx = ctx
        self.parent = parent
        self._declarations = collections.OrderedDict()
        self._functions = collections.OrderedDict()
        self._type_cache = {}

    def __repr__(self) -> str:
        return f"<Scope parent={self.parent} at 0x{id(self):x}>"

    @property
    def declarations(self) -> t.AbstractSet[IdentifierDeclaration]:
        """
        Variable and constant declarations of the scope.
        """
        return frozenset(self._declarations.values())

    @property
    def variable_declarations(self) -> t.Sequence[VariableDeclaration]:
        """
        Variable declarations of the scope.
        """
        return tuple(
            declaration
            for declaration in self._declarations.values()
            if isinstance(declaration, VariableDeclaration)
        )

    @property
    def constant_declarations(self) -> t.Sequence[ConstantDeclaration]:
        """
        Constant declarations of the scope.
        """
        return tuple(
            declaration
            for declaration in self._declarations.values()
            if isinstance(declaration, ConstantDeclaration)
        )

    @property
    def clock_declarations(self) -> t.AbstractSet[VariableDeclaration]:
        """
        Variable declarations of clock variables of the scope.
        """
        # FIXME: this does not return declarations with a bounded CLOCK type
        return frozenset(
            declaration
            for declaration in self._declarations.values()
            if (
                isinstance(declaration, VariableDeclaration)
                and declaration.typ == types.CLOCK
            )
        )

    def create_child_scope(self) -> Scope:
        """
        Creates a child scope.
        """
        return Scope(self.ctx, parent=self)

    def get_type(self, expr: expressions.Expression) -> types.Type:
        """
        Returns the (inferred) type of the given expression in the scope.
        """
        if expr not in self._type_cache:
            inferred_type = expr.infer_type(self)
            inferred_type.validate_in(self)
            self._type_cache[expr] = inferred_type
        return self._type_cache[expr]

    def get_function(self, name: str) -> functions.FunctionDefinition:
        """
        Retrieves a :class:`FunctionDefinition` by its name.

        Raises :class:`~errors.NotFoundError` if no such definition exists.
        """
        try:
            return self._functions[name]
        except KeyError:
            if self.parent is None:
                raise errors.NotFoundError(f"no function with name {name} found")
            return self.parent.get_function(name)

    def is_local(self, identifier: str) -> bool:
        """
        Checks whether the identifier is locally declared in the scope.
        """
        return identifier in self._declarations

    def is_declared(self, identifier: str) -> bool:
        """
        Checks whether the identifier is declared in the scope.
        """
        if identifier in self._declarations:
            return True
        if self.parent is not None:
            return self.parent.is_declared(identifier)
        return False

    def get_scope(self, identifier: str) -> Scope:
        """
        Retrieves the scope in which the given identifier is declared.

        Raises :class:`~errors.NotFoundError` if no such identifier is declared.
        """
        if identifier in self._declarations:
            return self
        else:
            if self.parent is None:
                raise errors.NotFoundError(
                    f"identifier {identifier!r} is unbound in scope {self!r}"
                )
            return self.parent.get_scope(identifier)

    def lookup(self, identifier: str) -> IdentifierDeclaration:
        """
        Retrieves the declaration for the given identifier.

        Raises :class:`~errors.NotFoundError` if no such identifier is declared.
        """
        try:
            return self._declarations[identifier]
        except KeyError:
            if self.parent is None:
                raise errors.NotFoundError(
                    f"identifier {identifier!r} is unbound in scope {self!r}"
                )
            return self.parent.lookup(identifier)

    def add_declaration(
        self, declaration: IdentifierDeclaration, *, validate: bool = True
    ) -> None:
        """
        Adds an identifier declaration to the scope.

        The flag `validate` specifies whether the declaration should
        be validated within the scope before adding it. In case
        validation fails, a :class:`~errors.ModelingError` is raised.

        Raises :class:`~errors.ModelingError` in case the identifier
        has already been declared.
        """
        if declaration.identifier in self._declarations:
            raise errors.InvalidDeclarationError(
                f"identifier {declaration.identifier!r} has already been declared"
            )
        if validate:
            declaration.validate(self)
        self._declarations[declaration.identifier] = declaration

    def declare_variable(
        self,
        identifier: str,
        typ: types.Type,
        *,
        is_transient: t.Optional[bool] = None,
        initial_value: t.Optional[expressions.ValueOrExpression] = None,
        comment: t.Optional[str] = None,
    ) -> None:
        """
        Declares a variable within the scope.

        The parameters are passed to :class:`VariableDeclaration` with
        the exception of `initial_value`. When provided with a value
        which is not an expressions, this function implicitly converts
        the provided value into an expression using :func:`ensure_expr`.

        Raises :class:`~errors.ModelingError` in case the identifier
        has already been declared.
        """
        value = None
        if initial_value is not None:
            value = expressions.ensure_expr(initial_value)
        self.add_declaration(
            VariableDeclaration(
                identifier,
                typ,
                is_transient=is_transient,
                initial_value=value,
                comment=comment,
            )
        )

    def declare_constant(
        self,
        identifier: str,
        typ: types.Type,
        *,
        value: t.Optional[expressions.ValueOrExpression] = None,
        comment: t.Optional[str] = None,
    ) -> None:
        """
        Declares a constant within the scope.

        The parameters are passed to :class:`ConstantDeclaration` with
        the exception of `value`. When provided with a value which is
        not an expressions, this function implicitly converts the
        provided value into an expression using :func:`ensure_expr`.

        Raises :class:`~errors.ModelingError` in case the identifier
        has already been declared.
        """
        if value is None:
            self.add_declaration(
                ConstantDeclaration(identifier, typ, comment=comment, value=None)
            )
        else:
            self.add_declaration(
                ConstantDeclaration(
                    identifier,
                    typ,
                    comment=comment,
                    value=expressions.ensure_expr(value),
                )
            )

    def define_function(
        self,
        name: str,
        parameters: t.Sequence[functions.FunctionParameter],
        returns: types.Type,
        body: expressions.Expression,
    ) -> functions.FunctionDefinition:
        """
        Defines a function within the scope.

        The parameters are passed to :class:`FunctionDefinition`.

        Raises :class:`~errors.ModelingError` in case an identically
        named function already exists.
        """
        if name in self._functions:
            raise errors.ModelingError(f"a function named {name!r} already exists")
        definition = functions.FunctionDefinition(
            name, tuple(parameters), returns, body
        )
        self._functions[name] = definition
        return definition


class Context:
    """
    Represents a *modeling context*.

    Attributes:
        model_type:
            The :class:`ModelType` of the modeling context.
        global_scope:
            The global :class:`Scope` of the modeling context.
    """

    model_type: ModelType
    global_scope: Scope

    _automata: t.Set[Automaton]
    _networks: t.Set[Network]

    _action_types: t.Dict[str, actions.ActionType]
    _named_properties: t.Dict[str, PropertyDefinition]

    _metadata: t.Dict[str, str]

    def __init__(self, model_type: ModelType) -> None:
        self.model_type = model_type
        self.global_scope = Scope(self)
        self._automata = set()
        self._networks = set()
        self._action_types = {}
        self._named_properties = {}
        self._metadata = {}

    def __repr__(self) -> str:
        return f"<Context model_type={self.model_type} at 0x{id(self):x}>"

    @property
    def automata(self) -> t.AbstractSet[Automaton]:
        """
        The set of automata defined on the modeling context.
        """
        return self._automata

    @property
    def networks(self) -> t.AbstractSet[Network]:
        """
        The set of networks defined on the modeling context.
        """
        return self._networks

    @property
    def metadata(self) -> t.Mapping[str, str]:
        """
        Additional metadata associated with the modeling
        context (e.g., author information).
        """
        return self._metadata

    @property
    def action_types(self) -> t.Mapping[str, actions.ActionType]:
        """
        The action types defined on the modeling context.
        """
        return self._action_types

    @property
    def properties(self) -> t.Mapping[str, PropertyDefinition]:
        """
        The properties defined on the modeling context.
        """
        return self._named_properties

    def update_metadata(self, metadata: t.Mapping[str, str]) -> None:
        """
        Updates the metadata with the provided mapping.
        """
        self._metadata.update(metadata)

    def get_automaton_by_name(self, name: str) -> Automaton:
        """
        Retrieves an automaton by its name.

        Raises :class:`~errors.NotFoundError` if no such automaton exists.
        """
        for automaton in self._automata:
            if automaton.name == name:
                return automaton
        raise errors.NotFoundError(f"there exists no automaton named {name!r}")

    def get_network_by_name(self, name: str) -> Network:
        """
        Retrives a network by its name.

        Raises :class:`~errors.NotFoundError` if no such network exists.
        """
        for network in self._networks:
            if network.name == name:
                return network
        raise errors.NotFoundError(f"there exists no network named {name!r}")

    def get_property_definition_by_name(self, name: str) -> PropertyDefinition:
        """
        Retrieves a property definition by its name.

        Raises :class:`~errors.NotFoundError` if no
        such property definition exists.
        """
        try:
            return self._named_properties[name]
        except KeyError:
            raise errors.NotFoundError(
                f"there exists no property definition named {name!r}"
            )

    def get_action_type_by_name(self, name: str) -> actions.ActionType:
        """
        Retrives an action type by its name.

        Raises :class:`~errors.NotFoundError` if no such action type exists.
        """
        try:
            return self._action_types[name]
        except KeyError:
            raise errors.NotFoundError(f"there exists no action type named {name!r}")

    def _add_action_type(self, action_type: actions.ActionType) -> None:
        """
        Adds an action type to the modeling context.

        Raises :class:`~errors.ModelingError` if an identically
        named action type already exists.
        """
        if action_type.label in self._action_types:
            raise errors.ModelingError(
                f"an action type with name {action_type.label!r} already exists"
            )
        self._action_types[action_type.label] = action_type

    def create_action_type(
        self, name: str, *, parameters: t.Sequence[actions.ActionParameter] = ()
    ) -> actions.ActionType:
        """
        Creates a new action type with the given name and parameters.

        Raises :class:`~errors.ModelingError` if an identically
        named action type already exists.
        """
        if name in self._action_types:
            raise errors.ModelingError(f"action type with name {name!r} already exists")
        action_type = actions.ActionType(name, tuple(parameters))
        self._add_action_type(action_type)
        return action_type

    def create_automaton(self, *, name: t.Optional[str] = None) -> Automaton:
        """
        Creates an automaton with the given optional name and returns it.
        """
        automaton = Automaton(self, name=name)
        self._automata.add(automaton)
        return automaton

    def create_network(self, *, name: t.Optional[str] = None) -> Network:
        """
        Creates a network with the given optional name and returns it.
        """
        network = Network(self, name=name)
        self._networks.add(network)
        return network

    def define_property(
        self, name: str, expression: expressions.Expression
    ) -> PropertyDefinition:
        """
        Defines a property on the modeling context.
        """
        definition = PropertyDefinition(name, expression)
        self._named_properties[name] = definition
        return definition



# -*- coding:utf-8 -*-
#
# Copyright (C) 2019-2020, Maximilian Köhl <koehl@cs.uni-saarland.de>

from __future__ import annotations

import typing as t

import dataclasses


JSON = t.Union[None, int, float, str, t.Sequence["JSON"], t.Mapping[str, "JSON"]]


@dataclasses.dataclass(frozen=True)
class JSONObject:
    parent: t.Optional[JSONObject]
    key: t.Union[int, str, None]


@dataclasses.dataclass(frozen=True)
class JSONPrimitive(JSONObject):
    value: t.Union[None, int, float, str]


class JSONSequence(JSONObject, t.Sequence[JSON]):
    pass


class JSONMapping(t.Mapping[str, JSON]):
    pass
